"""
MPI enabled data parallel optimizers
"""

import inspect
import math
import torch
import torch.distributed
from torch.nn.parallel import DistributedDataParallel as tDDP
from typing import Union, List, Tuple, Dict

from ..core.communication import MPICommunication
from ..core.communication import MPI
from ..core.communication import MPI_WORLD
from .utils import DetectMetricPlateau


__all__ = ["DataParallelOptimizer", "DASO"]


def __sum_f16_cb(buffer_a, buffer_b, _):
    # MPI custom sum function to use torch.half
    tens_a = torch.HalfTensor().set_(torch.HalfStorage.from_buffer(buffer_a, "native"))
    tens_b = torch.HalfTensor().set_(torch.HalfStorage.from_buffer(buffer_b, "native"))
    tens_b += tens_a
    nelem = torch.prod(torch.tensor(tens_b.shape)).item()
    new_buff = MPI.memory.fromaddress(tens_b.data_ptr(), nbytes=tens_b.element_size() * nelem)
    buffer_b[:] = new_buff


def __sum_bfloat_cb(buffer_a, buffer_b, _):
    # MPI custom sum function to use torch.bfloat16
    tens_a = torch.BFloat16Tensor().set_(torch.BFloat16Storage.from_buffer(buffer_a, "native"))
    tens_b = torch.BFloat16Tensor().set_(torch.BFloat16Storage.from_buffer(buffer_b, "native"))
    tens_b += tens_a
    nelem = int(tens_b.numel())
    new_buff = MPI.memory.fromaddress(tens_b.data_ptr(), nbytes=nelem * tens_b.element_size())
    buffer_b[:] = new_buff


# create new MPI OPs
mpi_sum_f16 = MPI.Op.Create(__sum_f16_cb, commute=True)
mpi_sum_bfloat = MPI.Op.Create(__sum_bfloat_cb, commute=True)


class DASO:
    r"""
    Optimizer wrapper to use the Distributed Asynchronous and Selective Optimization (DASO) method.

    This optimizer uses a local torch optimizer combined with the :func:`nn.DataParallelMultiGPU <heat.nn.data_parallel.DataParallelMultiGPU>`
    to create local DPNNs on each node consisting of the GPUs on each node. Then those networks communicate
    globally with MPI groups, each of which has a single GPU on each node.

    DASO uses both local and global synchronization operations. Local synchronization operations are intended to be
    done very frequently while global synchronizations are conducted asynchronously as the next batches are
    computed.

    This implementation requires that all nodes have the name number of GPUs.

    There are four phases to training:

        1. initialization: steps 1 to 8 below
        2. Warmup phase: blocking averaging update occurs for global synchronization step
        3. Cycling phase: for the global synchronization, the data is sent after a number of batches. the number of batches between synchronizations is referred to as `global_skips`. After the data is sent a number of batches pass before it is received (`batches_to_wait`). both of these cycle downward from `max_global_skips` for the global skips and 1/4th this value for `batches_to_wait`. When both values are equal to 1 and the loss is stable it will be reset to the initial values, then will decay again.
        4. Cooldown phase: blocking averaging update occurs for global synchronization step

    As example usage of this can be found in `heat/examples/nn/imagenet-DASO.py <https://github.com/helmholtz-analytics/heat/blob/504-docstring-formatting/examples/nn/imagenet-DASO.py>`_.

    The recommended checklist for using this class is as follows:

        1. initialize the local PyTorch process group and set the default device of the local GPUs.
        2. define the torch network
        3. define the `local_optimizer` -> a torch optimizer of your choice (tested with SGD)
        4. optional, choose a learning rate scheduler. This is only for those learning rates which will also step the optimizer
        5. initialize DASO with the local optimizers and parameters
        6. initialize :func:`nn.DataParallelMultiGPU <heat.nn.data_parallel.DataParallelMultiGPU>` with the torch network and DASO
        7. If using automatic mixed precision (:class:`torch.cuda.amp`), initialize the gradient scaler and add it to DASO (:func:`add_scaler`)
        8. ensure that the DataLoaders evenly distribute the data between all the processes. This can be done by using the `torch.utils.data.distributed.DistributedSampler <https://pytorch.org/docs/stable/data.html#torch.utils.data.distributed.DistributedSampler>`_ with the `num_replicas` and `rank` parameters
        9. call `daso_optimizer.epoch_loss_logic(training_loss)` at the end of
        10. set the number of batches per epoch (`daso_optimizer.last_batch = number_of_batches`)
        11. ensure that the step function used in training is that of the DASO optimizer

    Parameters
    ----------
    local_optimizer: torch.optim.Optimizer
        This optimizer handles the optimization of the local NN. Example: `torch.optim.SGD`. \n
        This can be any optimizer, although tests were only completed with SGD. Other optimizers may show
        unexpected behavior.
    total_epochs: int
        The total number of epochs for training. Needed to determine when to enter the cooldown phase.
    comm: MPICommunication, optional
        The MPI communicator to use for training. \n
        Default: :func:`MPI_WORLD <heat.core.comm.MPI_WORLD>`
    warmup_epochs: int, optional
        The number of epochs to complete with a blocking averaging operation after each batch before entering
        the cycling phase.\n
        Default: 4
    cooldown_epochs: int, optional
        The number of epochs with blocking averaging operations after each batch at the end of training.\n
        Default: 4
    scheduler: torch.optim.lr_scheduler, optional
        Local PyTorch learning rate scheduler. This must be used in the case that the scheduler's `step` function
        is supposed to be called instead of the optimizer's `step` function.\n
        Default: None
    stability_level: float, optional
        This can be viewed as the percent change threshold that the loss must exceed to be judged as improving.
        When the loss is within this percent change for 2 epochs, then it is judged as stable.\n
        Default: 0.05
    max_global_skips: int, optional
        The maximum number of batches between the beginning of a global synchronization process.\n
        Default: 8
    sending_chunk_size: int, optional
        During the global synchronization step, the network parameters are split into chunks of data to overlap
        communication and computation. This value is the maximum chunk size.\n
        Default: 10,000,000
    downcast_type: torch.dtype, optional
        Options: [torch.bfloat16, torch.half, torch.float]
        When the network parameters are sent during the global synchronization step, they are cast down to
        a smaller dtype, by default this is `torch.bfloat16`. Smaller torch dtypes are not implemented.
        torch.bfloat16.\n
        Default: torch.bfloat16
    use_mpi_groups: bool, optional
        Use MPI groups to divide the global communicator. If True, use MPI GROUPs, otherwise, use MPI SPLIT.\n
        Default: True
    skip_reduction_factor: int, optional
        How much to reduce the global/local skips by when the loss has stabilized.\n
        Default: 2
    local_skip_factor: int, optional
        How many local skips occur per global skip, i.e. number of local skips = global_skips // local_skip_factor.\n
        Default: 4
    verbose: bool, optional
        If true, print out a collection of debug messages.\n
        Default: False
    """

    def __init__(
        self,
        local_optimizer: torch.optim.Optimizer,
        total_epochs: int,
        comm: MPICommunication = MPI_WORLD,
        warmup_epochs: int = 4,
        cooldown_epochs: int = 4,
        scheduler: torch.optim.lr_scheduler = None,
        stability_level: float = 0.05,
        max_global_skips: int = 8,
        sending_chunk_size: int = 10_000_000,
        downcast_type: torch.dtype = torch.bfloat16,
        use_mpi_groups: bool = True,
        skip_reduction_factor: int = 2,
        local_skip_factor: int = 4,
        verbose: bool = False,
    ):  # noqa: D107
        # check dtypes
        frame = inspect.currentframe()
        init_args = inspect.getargvalues(frame)[3]
        self.__init_checktypes(init_args)

        self.cast_dtype = downcast_type
        if downcast_type == torch.bfloat16:
            self.cast_fn = mpi_sum_bfloat
        elif downcast_type == torch.half:
            self.cast_fn = mpi_sum_f16
        else:
            self.cast_fn = MPI.SUM

        self.comm = comm
        self.verbose = verbose
        self.local_optimizer = local_optimizer
        self.params_ref = local_optimizer.param_groups[0]["params"]
        # reference of optimizer's params
        self.scheduler = scheduler

        rank = self.comm.rank
        loc_gpus = torch.cuda.device_count()
        # this assumes that there are an equal number of GPUs per node,
        #   if a change is desired a comm her to find the lowest number would work for this, however
        #   a change would also need to be made in heat.nn.DataParallelMultiGPU
        self.loc_gpus = loc_gpus
        local_rank = rank % loc_gpus
        self.local_skip = 1
        if loc_gpus > 1:
            base_loc_ranks = list(range(0, self.comm.size, loc_gpus))
            reduced_comms, reduced_ranks = [], []
            for i in range(loc_gpus):
                lp_ranks = [j + i for j in base_loc_ranks]
                if use_mpi_groups:
                    new_group = self.comm.group.Incl(lp_ranks)
                    new_comm = self.comm.Create_group(new_group)
                    reduced_comms.append(MPICommunication(new_comm))
                else:
                    color = 111 + i if rank in lp_ranks else 222 + i
                    key = 0 + i if rank in lp_ranks else 444 + i
                    reduced_comms.append(MPICommunication(self.comm.Split(color, key)))
                reduced_ranks.append(tuple(lp_ranks))
            self.reduced_comms, self.reduced_ranks = reduced_comms, reduced_ranks
            self.base_loc_ranks = base_loc_ranks
            self.device = f"cuda:{str(local_rank)}"
            torch.cuda.set_device(device=self.device)

        self.current_batch, self.last_batch = 0, None

        self._prev_params = []
        self.epoch = 0
        self._send_mod, self._send_mod_m1 = 0, None

        self.global_skip = 0
        self.local_skip = 0
        self.batches_to_wait = 0
        self.max_gs = max_global_skips

        self.warmup_epochs = warmup_epochs
        self.cooldown_epochs = cooldown_epochs
        self.total_epochs = total_epochs

        # used in the sending of the params
        self._param_send_buffer_size = None
        self.param_dict, self.shapes = None, None
        self._param_send_shp = None
        self.split = None

        self.skip_reduction_factor = skip_reduction_factor
        # the local_skip_factor is the factor by which the global skips are divided by initially
        # and upon reset
        self.local_skip_factor = local_skip_factor

        self.stability = DetectMetricPlateau(patience=2, threshold=stability_level)

        self._gs8_waits = 3
        self._gs8_waited = 0

        self.split_val = sending_chunk_size

        # TODO: its possible that the split indexes could be used to avoid the concatenating method used currently
        self.split_inds = None
        self.amp = False
        self.print0("Finished DASO init")

    def add_scaler(self, scaler: torch.cuda.amp.GradScaler) -> None:
        """
        Create a reference to torch's `torch.cuda.amp.GradScaler <https://pytorch.org/docs/stable/notes/amp_examples.html>`_ used in torch's automatic mixed
        precision.

        Parameters
        ----------
        scaler: torch.cuda.amp.GradScaler
            the gradient scaler to be used
        """
        self.scaler = scaler
        self.amp = True

    @staticmethod
    def __init_checktypes(args: Dict) -> None:
        # this does all of the checks and raises for the parameters for init
        if not isinstance(args["local_optimizer"], torch.optim.Optimizer):
            raise TypeError(
                f"Local optimizer must be a torch optimizer object, currently {type(args['local_optimizer'])}"
            )
        if not isinstance(args["comm"], MPICommunication):
            raise TypeError(
                f"Comm object must be a ht.MPICommunication object, currently {type(args['comm'])}"
            )
        if not isinstance(args["total_epochs"], int):
            raise TypeError(f"total_epochs must be an int, currently {type(args['total_epochs'])}")
        if not isinstance(args["warmup_epochs"], int):
            raise TypeError(
                f"warmup_epochs must be an int, currently {type(args['warmup_epochs'])}"
            )
        if not isinstance(args["cooldown_epochs"], int):
            raise TypeError(
                f"cooldown_epochs must be an int, currently {type(args['cooldown_epochs'])}"
            )
        if args["scheduler"] is not None and not issubclass(
            args["scheduler"], torch.optim.lr_scheduler._LRScheduler
        ):
            raise TypeError(
                f"scheduler must be a torch learning rate scheduler, currently {args['scheduler']}"
            )
        if not isinstance(args["stability_level"], float):
            raise TypeError(
                f"stability_level must be a float, currently {type(args['stability_level'])}"
            )
        if not isinstance(args["max_global_skips"], int):
            raise TypeError(
                f"max_global_skips must be an int, currently {type(args['max_global_skips'])}"
            )
        if not isinstance(args["sending_chunk_size"], int):
            raise TypeError(
                f"sending_chunk_size must be an int, currently {type(args['sending_chunk_size'])}"
            )
        if not isinstance(args["verbose"], bool):
            raise TypeError(f"verbose must be a bool, currently {type(args['verbose'])}")
        if not isinstance(args["use_mpi_groups"], bool):
            raise TypeError(
                f"`use_mpi_grus` must be a bool, currently {type(args['use_mpi_groups'])}"
            )
        if not isinstance(args["downcast_type"], torch.dtype):
            raise TypeError(
                f"downcast_type must be a torch.dtype, currently {args['downcast_type']}"
            )
        if args["downcast_type"] not in [torch.bfloat16, torch.half, torch.float]:
            raise ValueError(
                f"downcast_type must be one of [torch.bfloat16, torch.half, torch.float], "
                f"currently {args['downcast_type']}"
            )
        if not isinstance(args["skip_reduction_factor"], int):
            raise TypeError(
                f"skip_reduction_factor must be an integer, currently {type(args['skip_reduction_factor'])}"
            )
        if not isinstance(args["local_skip_factor"], int):
            raise TypeError(
                f"local_skip_factor must be an integer, currently {type(args['local_skip_factor'])}"
            )

        if args["warmup_epochs"] < 0:
            raise ValueError(f"warmup_epochs must be >= 0, currently {args['warmup_epochs']}")
        if args["cooldown_epochs"] < 0:
            raise ValueError(f"cooldown_epochs must be >= 0, currently {args['cooldown_epochs']}")
        if args["max_global_skips"] < 0:
            raise ValueError(f"stablitiy_level must be >= 0, currently {args['max_global_skips']}")
        if args["sending_chunk_size"] <= 0:
            raise ValueError(
                f"sending_chunk_size must be > 0, currently {args['sending_chunk_size']}"
            )
        if args["total_epochs"] <= 0:
            raise ValueError(f"total_epochs must be > 0, currently {args['total_epochs']}")
        if args["skip_reduction_factor"] <= 0:
            raise ValueError(
                f"skip_reduction_factor must be > 0, currently {args['skip_reduction_factor']}"
            )
        if args["local_skip_factor"] <= 0:
            raise ValueError(
                f"local_skip_factor must be > 0, currently {args['local_skip_factor']}"
            )

    @torch.no_grad()
    def epoch_loss_logic(
        self, loss: Union[torch.Tensor, int, float], loss_globally_averaged: bool = False
    ) -> None:
        """
        Function controlling the number of batches between global synchronizations and the batches to wait before
        receiving the sent parameters. The warm-up and cool-down phases are also controlled here.

        This function should be called at the end of each epoch with the training loss value at the end of the epoch.

        The number of batches between local synchronizations can also be modified here with minor code adjustments.

        Parameters
        ----------
        loss: torch.Tensor or float
            loss value of the current epoch
        loss_globally_averaged: bool, optional
            boolean if the loss is already globally averaged
        """
        if not loss_globally_averaged:
            loss_send = torch.zeros(self.comm.size)
            # loss.data -> this will get the raw number from the lass value and nothing else
            loss_send[self.comm.rank] = loss.data if isinstance(loss, torch.Tensor) else loss
            self.comm.Allreduce(MPI.IN_PLACE, loss_send, MPI.SUM)
            avg_loss = torch.mean(loss_send)
        else:
            avg_loss = torch.tensor(loss)

        if self.epoch < self.warmup_epochs:
            self.global_skip = 0
            self.local_skip = 0
            self.batches_to_wait = 0

            self.print0(
                f"Warmup Phase, parameters of next epoch\n\t Global Skips: {self.global_skip}, "
                f" Local Skips {self.local_skip}, Batches to wait: {self.batches_to_wait}"
            )
            return

        elif self.warmup_epochs == self.epoch:
            self.global_skip = 4
            self.local_skip = 1
            self.batches_to_wait = 1

            self.print0(
                f"End of Warmup Phase, parameters of next epoch\n\t Global Skips: {self.global_skip}, "
                f" Local Skips {self.local_skip}, Batches to wait: {self.batches_to_wait}"
            )

        if self.epoch >= self.total_epochs - self.cooldown_epochs:
            self.global_skip = 0
            self.local_skip = 0
            self.batches_to_wait = 0

            self.print0(
                f"Cooldown Phase, parameters of next epoch:\n\tGlobal Skips: {self.global_skip}, "
                f" Local Skips {self.local_skip:.4f}, Batches to wait: {self.batches_to_wait}"
            )
            return

        if self.global_skip == self.max_gs and self.max_gs > 4:
            self._gs8_waited += 1

        self.print0(
            f"Best loss value: {self.stability.best * (1.0 - self.stability.threshold):.4f}"
            f" Current loss: {avg_loss:.4f}, Worse epochs: {self.stability.num_bad_epochs}"
        )

        stable = self.stability.test_if_improving(avg_loss)

        if stable and self.global_skip > 1:
            # drop gs by factor of 2
            self.global_skip //= self.skip_reduction_factor
            self.local_skip //= self.skip_reduction_factor
            self.batches_to_wait -= 1  # old was //= 2

            self.print0("dropping skips")

            if self.global_skip > 0:
                if self.batches_to_wait == 0:
                    self.batches_to_wait = 1
                if self.local_skip == 0:
                    self.local_skip = 1
            self._gs8_waited = 0
        elif self.global_skip == 1 and stable:
            self.global_skip = self.max_gs
            self.local_skip = self.max_gs // self.local_skip_factor
            self.batches_to_wait = self.max_gs // self.local_skip_factor

            self._gs8_waited = 0
        self.print0(
            f"\tNext Parameters: Global Skips: {self.global_skip}, Local Skips {self.local_skip}, "
            f" Batches to wait: {self.batches_to_wait}, \n\tCurrent loss: {avg_loss:.4f}, "
            f" Worse epochs: {self.stability.num_bad_epochs}"
        )

    @torch.no_grad()
    def _global_sync(self, batches_to_wait: int) -> None:
        """
        Performs a global synchronization. If `batches_to_wait > 0` this will wait for that many
        batches before received in the parameters.

        Full syncs are only performed on a single MPI group
        """
        current_comm = self.reduced_comms[self._send_mod]
        current_ranks = self.reduced_ranks[self._send_mod]

        if self.comm.rank in current_ranks:
            self._gs_send_params(current_comm, batches_to_wait)

        if self.batches_to_wait != 0:
            # update parameters from the last sending (if there)
            self._gs_rcv_update_params()  # -> splits off irrelevant ranks
            # needs to happen on all ranks:
            self._local_update(self._send_mod_m1)

        if self.current_batch == self.last_batch or self.batches_to_wait == 0:
            # receive the sent data to sync params across all ranks
            if self.comm.rank in current_ranks:
                self._gs_rcv_update_params_last_batch(current_ranks)
            else:
                if len(self._prev_params) > 0:
                    raise ValueError(
                        f"DEBUG: OFF RANKS! len(prev_params) > 0! {len(self._prev_params)}"
                        f" batch number {self.current_batch}"
                    )
            self._local_update(self._send_mod)

            self._send_mod_m1 = None

            if self.current_batch == self.last_batch:
                self._send_mod = 0
                self.epoch += 1
                self.current_batch = 0
            else:
                self.current_batch += 1
                self._send_mod = self._send_mod + 1 if self._send_mod <= self.loc_gpus - 2 else 0
        else:
            self.current_batch += 1
            self._send_mod_m1 = self._send_mod
            self._send_mod = self._send_mod + 1 if self._send_mod <= self.loc_gpus - 2 else 0

    def _gs_create_param_dict(self) -> Tuple[Dict, Dict]:
        """
        Create the shape and param dictionary used for sending parameters around the MPI world.
        this will also define the buffer size if it was not previously defined.
        """
        if self.shapes is not None:
            return self.param_dict, self.shapes
        param_dict = {}
        shapes = {}
        st = 0
        for name, param in self.module.named_parameters():
            param_dict[name] = param
            numel = param.numel()
            shapes[name] = [param.shape, slice(st, st + numel), param.dtype]
            st += numel

        if self._param_send_buffer_size is None:
            # use the total number of elements to define the sending buffer shape (single int)
            self._param_send_buffer_size = st

        self.param_dict = param_dict
        self.shapes = shapes
        return param_dict, shapes

    @torch.no_grad()
    def _gs_rcv_update_params(self) -> None:
        """
        Receive the previously sent parameters for the last sending MPI group.
        this is also where the sent and local parameters are merged together.
        """
        # wait for the global sync data and update on the selected rank,
        if self._send_mod_m1 is None:
            return
        prev_ranks = self.reduced_ranks[self._send_mod_m1]
        if self.comm.rank not in prev_ranks or len(self._prev_params) == 0:
            # if no old gradients, return without doing anything
            return

        prev_params = self._prev_params.pop(0)
        batches_between = float(prev_params[3])
        shapes = prev_params[2]
        # add the weighted average to the received params
        numer = batches_between * 2.0 if batches_between > 0.0 else 1.0
        denom = float(len(prev_ranks) + numer)
        factor = numer / denom
        if not self.split:
            # only a single buffer
            prev_params[0].Wait()
            rcv_params = prev_params[1] / denom
            for name, param in self.module.named_parameters():
                if param.requires_grad:
                    update = (
                        rcv_params[shapes[name][1]].reshape(shapes[name][0]).to(shapes[name][2])
                    )
                    # NOTE: update here is the sum of the params across the processes
                    param *= factor
                    param += update
        else:
            # receive the first buffer
            ind = 0
            prev_params[0][0].Wait()
            del prev_params[0][0]
            rcv_params = prev_params[1][ind] / denom
            # jit the parameter setting
            for name, param in self.module.named_parameters():
                if param.requires_grad:
                    if shapes[name][1].stop > len(rcv_params):
                        # when the end of the slice is higher than the amount received, wait for the next buffer
                        ind += 1
                        prev_params[0][0].Wait()
                        del prev_params[0][0]
                        new_rcv_params = prev_params[1][ind] / denom
                        rcv_params = torch.cat((rcv_params, new_rcv_params))
                    update = (
                        rcv_params[shapes[name][1]].reshape(shapes[name][0]).to(shapes[name][2])
                    )
                    # NOTE: update here is the sum of the params across the processes
                    param *= factor
                    param += update

    @torch.no_grad()
    def _gs_rcv_update_params_last_batch(self, current_ranks: Tuple) -> None:
        """
        Abstracted receive for the last batch (and if `global_skips` == 0)
        """
        if len(self._prev_params) > 1:
            raise ValueError(f"length of previous params > 1! {len(self._prev_params)}")
        prev_params = self._prev_params.pop(0)
        shapes = prev_params[2]
        if not self.split:
            prev_params[0].Wait()
            rcv_params = prev_params[1] / float(len(current_ranks))
            for name, param in self.module.named_parameters():
                if param.requires_grad:
                    param[:] = (
                        rcv_params[shapes[name][1]].reshape(shapes[name][0]).to(shapes[name][2])
                    )
        else:
            ind1 = 0
            prev_params[0][0].Wait()
            del prev_params[0][0]
            rcv_params = prev_params[1][ind1] / float(len(current_ranks))
            for name, param in self.module.named_parameters():
                if param.requires_grad:
                    while shapes[name][1].stop > len(rcv_params):
                        ind1 += 1
                        prev_params[0][0].Wait()
                        del prev_params[0][0]
                        new_rcv_params = prev_params[1][ind1] / float(len(current_ranks))
                        rcv_params = torch.cat((rcv_params, new_rcv_params))
                    param[:] = (
                        rcv_params[shapes[name][1]].reshape(shapes[name][0]).to(shapes[name][2])
                    )

    @torch.no_grad()
    def _gs_send_params(self, current_comm: MPICommunication, batches_to_wait: int) -> None:
        """
        Pack and send the data required for a global synchronization on the `current_comm` group

        `batches_to_wait` is sent with the parameters to keep track of this between sending and receiving
        """
        op = MPI.SUM
        cast = False
        cast_int = 2
        if self.global_skip < 1:
            # op = mpi_sum_bfloat
            cast = True
            op = self.cast_fn
            if self.cast_dtype == torch.bfloat16:
                cast_int = 0
            elif self.cast_dtype == torch.half:
                cast_int = 1
            # else:  # keep as floats (default case, see above)

        param_dict, shapes = self._gs_create_param_dict()
        sndparams = torch.zeros(
            self._param_send_buffer_size,
            device=self.device,
            dtype=self.cast_dtype if cast else None,
        )

        sndparams = self.__pack_data(sndparams, param_dict, cast_int)
        try:
            nans = sndparams.isnan().sum()
        except RuntimeError:
            # the isnan function isnt implemented in some cuda / torch implementations
            nans = sndparams.to(torch.half).isnan().sum()

        if nans:
            # check if there are NaNs, if so, stuff is bad
            raise ValueError(f"{nans} NaNs in `params` shit be fucked.")

        if not self.split and sndparams.numel() <= self.split_val:
            new_wait = current_comm.Iallreduce(MPI.IN_PLACE, sndparams, op)
            self._prev_params.append([new_wait, sndparams, shapes, batches_to_wait])
            return

        # if self.split or sndparams.numel() > self.split_val:
        self.split = True
        num_splits = math.ceil(len(sndparams) / self.split_val)
        splits = [self.split_val] * (num_splits - 1)
        rem = len(sndparams) - (self.split_val * (num_splits - 1))
        # first one will be smaller then the rest (the remainder is first)
        splits = [rem] + splits
        self.split_inds = splits
        params_list = [None] * num_splits
        prev = 0
        waits = [None] * num_splits
        for s in range(num_splits):
            # need to slice the params at the split points
            params_list[s] = sndparams[prev : splits[s] + prev]
            prev += splits[s]
            waits[s] = current_comm.Iallreduce(MPI.IN_PLACE, params_list[s], op)
        self._prev_params.append([waits, params_list, shapes, batches_to_wait])

    @torch.no_grad()
    def _local_update(self, sending_process: Tuple) -> None:
        # use torch to send the network parameters of a single process to the other processes
        if not torch.distributed.is_initialized() or sending_process is None:
            return
        snds = {
            name: torch.distributed.broadcast(param, sending_process, async_op=True)
            for name, param in self.module.named_parameters()
            if param.requires_grad
        }
        for name, param in self.module.named_parameters():
            if param.requires_grad:
                snds[name].wait()

    @staticmethod
    @torch.no_grad()
    @torch.jit.script
    def __pack_data(
        jtparams: torch.Tensor, iter_dict: Dict[str, torch.Tensor], cast: int
    ) -> torch.Tensor:
        """
        Jitted loop to pack the data into a flattened buffer to be sent
        """
        st = 0
        cast_type = torch.float if cast == 2 else torch.bfloat16 if cast == 0 else torch.half
        for par in iter_dict.values():
            if par.requires_grad:
                # flatten and prep the data for sending
                p = torch.flatten(par)
                p = p.to(cast_type)
                jtparams[st : st + par.numel()] = p
                st += par.numel()
        return jtparams

    def print0(self, *args, **kwargs) -> None:
        """
        Print a message on rank 0 if the class parameter `verbose` is set.
        """
        if self.comm.rank == 0 and self.verbose:
            print(*args, **kwargs)

    def reset(self) -> None:
        """
        Reset the optimizer to its base state
        """
        self.stability.reset()
        self.global_skip = 0
        self.local_skip = 0
        self.batches_to_wait = 0
        self.current_batch = 0
        self._prev_params = []
        self.epoch = 0
        self._gs8_waited = 0
        self.zero_grad()

    def set_model(self, model: torch.nn.Module) -> None:
        """
        Set the local model for the optimizer.
        This should be called during the init of :func:`nn.DataParallelMultiGPU <heat.nn.data_parallel.DataParallelMultiGPU>`.
        However, this can also be called manually.

        Parameters
        ----------
        model: torch.nn.Module
            the local torch model.
        """
        self.module = model

    def _start_local_sync(self) -> None:
        """
        *Start* local synchronizations for the next batches
        """
        if not isinstance(self.module, tDDP) or self.module.require_backward_grad_sync:
            # this has no effect if the module is not locally distributed in torch
            return
        self.module.require_backward_grad_sync = True

    def step(self) -> None:
        """
        Perform a single optimization step.
        This will perform the `step` operations of the local optimizer,
        local learning rate scheduler (if defined), and the gradient scaler used in automatic mixed
        precision (if defined).

        Also in the step is the logic used for when to send and receive the global/local synchronizations.
        Global Syncs occur on batches for which the modulus of the batch number and the `global_skip` number is 0.
        If `batches_to_wait` > 0, the next batches have only local syncs. After that number of batches,
        the data during the global sync phase is received.

        Local synchronization can also be turned off if desired by increasing `local_skips` above 1.

        Notes
        -----
        self.last_batch must be set!
        """
        if self.last_batch is None:
            raise ValueError(
                "self.last_batch must be set as the number of batches (len(dataloader))"
            )

        if self.amp:
            self.scaler.step(self.local_optimizer)
            # todo: add something to tell if the grads have infs or nans
            # Updates the scale for next iteration.
            self.scaler.update()
        elif self.scheduler is None:
            self.local_optimizer.step()
        else:
            self.scheduler.step()
        batch = self.current_batch
        # knowing next_batch is important to make sure that the local sync is on
        #       or if the next is the last batch
        next_batch = batch + 1
        gs = self.global_skip
        ls = self.local_skip
        # determine if to do the syncs
        gmod = batch % gs if gs > 0 else 0
        lmod = batch % ls if ls > 0 else 0

        batches_to_wait = self.batches_to_wait
        # ensure that the batch that will receive will be before the end of the training loop
        btw = (
            batches_to_wait
            if batches_to_wait + batch <= self.last_batch
            else self.last_batch - batch
        )
        # do full sync on global skips and on the last batch
        if batch == self.last_batch or gmod == 0:
            return self._global_sync(btw)

        if next_batch % gs == 0:
            self._start_local_sync()
            self.current_batch += 1
            return

        if gmod < btw:
            # do nothing on these batches (maintain the local sync)
            self.current_batch += 1
            if next_batch == self.last_batch:
                self._start_local_sync()
            return
        elif gmod == btw:
            # local updates should be on before this is called!
            self._gs_rcv_update_params()
            self._local_update(self._send_mod_m1)
            if ls > 1:
                self._stop_local_sync()

        if ls == 1 and next_batch != self.last_batch:
            self.current_batch += 1
            self._start_local_sync()
            return

        if lmod == 0:
            self._stop_local_sync()
        elif next_batch % ls == 0:
            self._start_local_sync()

        if next_batch == self.last_batch:
            self._start_local_sync()

        self.current_batch += 1

    def _stop_local_sync(self) -> None:
        """
        Stop local synchronizations for the next batches
        """
        if not isinstance(self.module, tDDP) or not self.module.require_backward_grad_sync:
            # this has no effect if the module is not locally distributed in torch
            return
        self.module.require_backward_grad_sync = False

    def zero_grad(self) -> None:
        """
        Reset gradients of local optimizer's parameters.
        """
        # reset view onto params in order to reset all gradients
        self.local_optimizer.param_groups[0]["params"] = self.params_ref[:]
        self.local_optimizer.zero_grad(set_to_none=False)


class DataParallelOptimizer:
    """
    Uses a torch.optim.Optimizer for data parallelism. It should be used in combination with DataParallel (DP) class.
    To optimize a DP module, DP optimizer has to be passed to DP module during its initialization.
    See :func:`nn.DataParallel <heat.nn.data_parallel.DataParallel>` for a basic example of usage.

    Attributes
    ----------
    torch_optimizer : torch.optim.Optimizer
        the wrapped Torch optimizer
    blocking : bool
        use blocking communications or not. will typically be overwritten by :func:`nn.DataParallel <heat.nn.data_parallel.DataParallel>`
    """

    def __init__(
        self, torch_optimizer: torch.optim.Optimizer, blocking: bool = False
    ):  # noqa: D107
        self.torch_optimizer = torch_optimizer
        if not isinstance(blocking, bool):
            raise TypeError(f"blocking parameter must be a boolean, currently {type(blocking)}")
        # flag indicating if communication during parameter updates is blocking.
        self.blocking_parameter_updates = blocking
        # flag indicating if optimizer should take a step during next iteration (only relevant for non-blocking)
        self.update_next = False
        # reference of optimizer's params
        self.params_ref = torch_optimizer.param_groups[0]["params"]

    def step(self) -> None:
        """
        Force torch optimizer to update model parameters. For blocking, optimizer immediately updates parameters. For
        non-blocking, optimizer will update parameters during next forward.
        """
        if self.blocking_parameter_updates:
            self.torch_optimizer.step()
        else:
            self.update_next = True

    def zero_grad(self) -> None:
        """
        Reset gradients of optimizer's params.
        """
        # reset view onto params in order to reset all gradients
        self.torch_optimizer.param_groups[0]["params"] = self.params_ref[:]
        self.torch_optimizer.zero_grad(set_to_none=False)



"""
File containing the heat.nn.functional submodule
"""

import sys
import torch


def func_getattr(name):
    """
    When a function is called for the heat.nn.functional module it will attempt to run the
    heat.nn.functional module with that name, then, if there is no such heat nn module,
    it will attempt to get the torch.nn.functional module of that name.
    """
    # call through to the torch.nn.functional module
    try:
        return torch.nn.functional.__getattribute__(name)
    except AttributeError:
        if name is not None:
            raise AttributeError("module not implemented in Torch Functional")



"""Implements the k-nearest neighbors (kNN) classifier"""
from typing import Callable

import heat as ht

from heat.core.dndarray import DNDarray


class KNeighborsClassifier(ht.BaseEstimator, ht.ClassificationMixin):
    """
    Implementation of the k-nearest-neighbors Algorithm [1].

    This algorithm predicts labels to data vectors by using an labeled training dataset as reference. The input vector
    to be predicted is compared to the training vectors by calculating the Euclidean distance between each of them. A
    majority vote of the k-nearest, i.e. closest or smallest distanced, training vectors labels is selected as
    predicted class.

    Parameters
    ----------
    n_neighbors : int, optional, default: 5
        Number of neighbours to consider when choosing label.
    effective_metric_ : Callable, optional
        The distance function used to identify the nearest neighbors, defaults to the Euclidean distance.

    References
    --------
    [1] T. Cover and P. Hart, "Nearest Neighbor Pattern Classification," in IEEE Transactions on Information Theory,
    vol. 13, no. 1, pp. 21-27, January 1967, doi: 10.1109/TIT.1967.1053964.
    """

    def __init__(self, n_neighbors: int = 5, effective_metric_: Callable = None):
        self.n_neighbors = n_neighbors
        self.effective_metric_ = (
            effective_metric_ if effective_metric_ is not None else ht.spatial.cdist
        )

        # init declaration to appease flake
        self.x = None
        self.y = None
        self.n_samples_fit_ = -1
        self.outputs_2d_ = True
        self.classes_ = None

    @staticmethod
    def one_hot_encoding(x: DNDarray) -> DNDarray:
        """
        One-hot-encodes the passed vector or single-column matrix.

        Parameters
        ----------
        x : DNDarray
            The data to be encoded.
        """
        n_samples = x.shape[0]
        n_features = ht.max(x).item() + 1

        one_hot = ht.zeros((n_samples, n_features), split=x.split, device=x.device, comm=x.comm)
        one_hot.lloc[range(one_hot.lshape[0]), x.larray] = 1

        return one_hot

    def fit(self, x: DNDarray, y: DNDarray):
        """
        Fit the k-nearest neighbors classifier from the training dataset.

        Parameters
        ----------
        x : DNDarray
            Labeled training vectors used for comparison in predictions, Shape=(n_samples, n_features).
        y : DNDarray
            Corresponding labels for the training feature vectors. Must have the same number of samples as ``x``.
            Shape=(n_samples) if integral labels or Shape=(n_samples, n_classes) if one-hot-encoded.

        Raises
        ------
        TypeError
            If ``x`` or ``y`` are not DNDarrays.
        ValueError
            If ``x`` and ``y`` shapes mismatch or are not two-dimensional matrices.

        Examples
        --------
        >>> samples = ht.rand(10, 3)
        >>> knn = KNeighborsClassifier(n_neighbors=1)
        >>> knn.fit(samples)
        """
        # check for type consistency
        if not isinstance(x, DNDarray) or not isinstance(y, DNDarray):
            raise TypeError(f"x and y must be DNDarrays but were {type(x)} {type(y)}")

        # ensure that x is a two-dimensional matrix
        if len(x.shape) != 2:
            raise ValueError(f"x must be two-dimensional, but was {len(x.shape)}")
        self.x = x
        self.n_samples_fit_ = x.shape[0]

        # ensure that x and y have the same number of samples
        if x.shape[0] != y.shape[0]:
            raise ValueError(
                f"Number of samples x and y samples mismatch, got {x.shape[0]}, {y.shape[0]}"
            )

        # checks the labels for correct dimensionality and encode one-hot
        if len(y.shape) == 1:
            self.y = self.one_hot_encoding(y)
            self.outputs_2d_ = False
        elif len(y.shape) == 2:
            self.y = y
            self.outputs_2d_ = True
        else:
            raise ValueError(f"y needs to be one- or two-dimensional, but was {len(y.shape)}")

    def predict(self, x: DNDarray) -> DNDarray:
        """
        Predict the class labels for the provided data.

        Parameters
        ----------
        x : DNDarray
            The test samples.
        """
        distances = self.effective_metric_(x, self.x)
        _, indices = ht.topk(distances, self.n_neighbors, largest=False)

        predictions = self.y[indices.flatten()]
        predictions.balance_()
        predictions = ht.reshape(predictions, (indices.gshape + (self.y.gshape[1],)))
        predictions = ht.sum(predictions, axis=1)

        self.classes_ = ht.argmax(predictions, axis=1)

        return self.classes_



"""Provides a collection of signal-processing operations"""

import torch
import numpy as np

from .communication import MPI
from .dndarray import DNDarray
from .types import promote_types
from .manipulations import pad, flip
from .factories import array, zeros
import torch.nn.functional as fc

__all__ = ["convolve"]


def convolve(a: DNDarray, v: DNDarray, mode: str = "full") -> DNDarray:
    """
    Returns the discrete, linear convolution of two one-dimensional `DNDarray`s or scalars.

    Parameters
    ----------
    a : DNDarray or scalar
        One-dimensional signal `DNDarray` of shape (N,) or scalar.
    v : DNDarray or scalar
        One-dimensional filter weight `DNDarray` of shape (M,) or scalar.
    mode : str
        Can be 'full', 'valid', or 'same'. Default is 'full'.
        'full':
          Returns the convolution at
          each point of overlap, with an output shape of (N+M-1,). At
          the end-points of the convolution, the signals do not overlap
          completely, and boundary effects may be seen.
        'same':
          Mode 'same' returns output  of length 'N'. Boundary
          effects are still visible. This mode is not supported for
          even-sized filter weights
        'valid':
          Mode 'valid' returns output of length 'N-M+1'. The
          convolution product is only given for points where the signals
          overlap completely. Values outside the signal boundary have no
          effect.

    Examples
    --------
    Note how the convolution operator flips the second array
    before "sliding" the two across one another:

    >>> a = ht.ones(10)
    >>> v = ht.arange(3).astype(ht.float)
    >>> ht.convolve(a, v, mode='full')
    DNDarray([0., 1., 3., 3., 3., 3., 2.])
    >>> ht.convolve(a, v, mode='same')
    DNDarray([1., 3., 3., 3., 3.])
    >>> ht.convolve(a, v, mode='valid')
    DNDarray([3., 3., 3.])
    >>> a = ht.ones(10, split = 0)
    >>> v = ht.arange(3, split = 0).astype(ht.float)
    >>> ht.convolve(a, v, mode='valid')
    DNDarray([3., 3., 3., 3., 3., 3., 3., 3.])

    [0/3] DNDarray([3., 3., 3.])
    [1/3] DNDarray([3., 3., 3.])
    [2/3] DNDarray([3., 3.])
    >>> a = ht.ones(10, split = 0)
    >>> v = ht.arange(3, split = 0)
    >>> ht.convolve(a, v)
    DNDarray([0., 1., 3., 3., 3., 3., 3., 3., 3., 3., 3., 2.], dtype=ht.float32, device=cpu:0, split=0)

    [0/3] DNDarray([0., 1., 3., 3.])
    [1/3] DNDarray([3., 3., 3., 3.])
    [2/3] DNDarray([3., 3., 3., 2.])
    """
    if np.isscalar(a):
        a = array([a])
    if np.isscalar(v):
        v = array([v])
    if not isinstance(a, DNDarray):
        try:
            a = array(a)
        except TypeError:
            raise TypeError(f"non-supported type for signal: {type(a)}")
    if not isinstance(v, DNDarray):
        try:
            v = array(v)
        except TypeError:
            raise TypeError(f"non-supported type for filter: {type(v)}")
    promoted_type = promote_types(a.dtype, v.dtype)
    a = a.astype(promoted_type)
    v = v.astype(promoted_type)

    if len(a.shape) != 1 or len(v.shape) != 1:
        raise ValueError("Only 1-dimensional input DNDarrays are allowed")
    if mode == "same" and v.shape[0] % 2 == 0:
        raise ValueError("Mode 'same' cannot be used with even-sized kernel")
    if not v.is_balanced():
        raise ValueError("Only balanced kernel weights are allowed")

    if v.shape[0] > a.shape[0]:
        a, v = v, a

    # compute halo size
    halo_size = torch.max(v.lshape_map[:, 0]).item() // 2

    # pad DNDarray with zeros according to mode
    if mode == "full":
        pad_size = v.shape[0] - 1
        gshape = v.shape[0] + a.shape[0] - 1
    elif mode == "same":
        pad_size = v.shape[0] // 2
        gshape = a.shape[0]
    elif mode == "valid":
        pad_size = 0
        gshape = a.shape[0] - v.shape[0] + 1
    else:
        raise ValueError(f"Supported modes are 'full', 'valid', 'same', got {mode}")

    a = pad(a, pad_size, "constant", 0)

    if a.is_distributed():
        if (v.lshape_map[:, 0] > a.lshape_map[:, 0]).any():
            raise ValueError(
                "Local chunk of filter weight is larger than the local chunks of signal"
            )
        # fetch halos and store them in a.halo_next/a.halo_prev
        a.get_halo(halo_size)
        # apply halos to local array
        signal = a.array_with_halos
    else:
        signal = a.larray

    # flip filter for convolution as Pytorch conv1d computes correlations
    v = flip(v, [0])
    if v.larray.shape != v.lshape_map[0]:
        # pads weights if input kernel is uneven
        target = torch.zeros(v.lshape_map[0][0], dtype=v.larray.dtype, device=v.larray.device)
        pad_size = v.lshape_map[0][0] - v.larray.shape[0]
        target[pad_size:] = v.larray
        weight = target
    else:
        weight = v.larray

    t_v = weight  # stores temporary weight

    # make signal and filter weight 3D for Pytorch conv1d function
    signal = signal.reshape(1, 1, signal.shape[0])
    weight = weight.reshape(1, 1, weight.shape[0])

    # cast to float if on GPU
    if signal.is_cuda:
        float_type = promote_types(signal.dtype, torch.float32).torch_type()
        signal = signal.to(float_type)
        weight = weight.to(float_type)
        t_v = t_v.to(float_type)

    if v.is_distributed():
        size = v.comm.size

        for r in range(size):
            rec_v = t_v.clone()
            v.comm.Bcast(rec_v, root=r)
            t_v1 = rec_v.reshape(1, 1, rec_v.shape[0])
            local_signal_filtered = fc.conv1d(signal, t_v1)
            # unpack 3D result into 1D
            local_signal_filtered = local_signal_filtered[0, 0, :]

            if a.comm.rank != 0 and v.lshape_map[0][0] % 2 == 0:
                local_signal_filtered = local_signal_filtered[1:]

            # accumulate filtered signal on the fly
            global_signal_filtered = array(
                local_signal_filtered, is_split=0, device=a.device, comm=a.comm
            )
            if r == 0:
                # initialize signal_filtered, starting point of slice
                signal_filtered = zeros(
                    gshape, dtype=a.dtype, split=a.split, device=a.device, comm=a.comm
                )
                start_idx = 0

            # accumulate relevant slice of filtered signal
            # note, this is a binary operation between unevenly distributed dndarrays and will require communication, check out _operations.__binary_op()
            signal_filtered += global_signal_filtered[start_idx : start_idx + gshape]
            if r != size - 1:
                start_idx += v.lshape_map[r + 1][0].item()
        return signal_filtered

    else:
        # apply torch convolution operator
        signal_filtered = fc.conv1d(signal, weight)

        # unpack 3D result into 1D
        signal_filtered = signal_filtered[0, 0, :]

        # if kernel shape along split axis is even we need to get rid of duplicated values
        if a.comm.rank != 0 and v.shape[0] % 2 == 0:
            signal_filtered = signal_filtered[1:]

        return DNDarray(
            signal_filtered.contiguous(),
            (gshape,),
            signal_filtered.dtype,
            a.split,
            a.device,
            a.comm,
            balanced=False,
        ).astype(a.dtype.torch_type())



"""
This file contains functions which may be useful for certain datatypes, but are not test in the heat framework
This file contains standalone utilities for data preparation which may be useful
The functions contained within are not tested, nor actively supported
"""

import base64
import numpy as np
import os
import struct


def dali_tfrecord2idx(train_dir, train_idx_dir, val_dir, val_idx_dir):
    """
    WARNING: This function likely requires adjustments and it is by no means a final product !!!
    this file contains standalone utilities for data preparation which may be useful
    this function contained within are not tested, nor actively supported

    prepare TFRecords indexes for use with DALI. It will produce indexes for all files in the
    given ``train_dir`` and ``val_dir`` directories
    """
    for tv in [train_dir, val_dir]:
        dir_list = os.listdir(tv)
        out = train_idx_dir if tv == train_dir else val_idx_dir
        for file in dir_list:
            with open(file, "rb") as f, open(out + file, "w") as idx:
                while True:
                    current = f.tell()
                    try:
                        # length
                        byte_len = f.read(8)
                        if len(byte_len) == 0:
                            break
                        # crc
                        f.read(4)
                        proto_len = struct.unpack("q", byte_len)[0]
                        # proto
                        f.read(proto_len)
                        # crc
                        f.read(4)
                        idx.write(str(current) + " " + str(f.tell() - current) + "\n")
                    except Exception:
                        print("Not a valid TFRecord file")
                        break


def merge_files_imagenet_tfrecord(folder_name, output_folder=None):
    """
    WARNING: This function likely requires adjustments and it is by no means a final product !!!
    this file contains standalone utilities for data preparation which may be useful
    this function contained within are not tested, nor actively supported

    merge multiple preprocessed imagenet TFRecord files together,
    result is one HDF5 file with all of the images stacked in the 0th dimension

    Parameters
    ----------
    folder_name : str, optional*
        folder location of the files to join, either filenames or folder_names must not be None
    output_folder : str, optional
        location to create the output files. Defaults to current directory

    Notes
    -----
    Metadata for both the created files (`imagenet_merged.h5` and `imagenet_merged_validation.h5`):

    The datasets are the combination of all of the images in the Image-net 2012 dataset.
    The data is split into training and validation.

    imagenet_merged.h5 -> training
    imagenet_merged_validation.h5 -> validation

    both files have the same internal structure:
    - file
            * "images" : encoded ASCII string of the decoded RGB JPEG image.
                    - to decode: `torch.as_tensor(bytearray(base64.binascii.a2b_base64(string_repr.encode('ascii'))), dtype=torch.uint8)`
                    - note: the images must be reshaped using: `.reshape(file["metadata"]["image/height"], file["metadata"]["image/height"], 3)`
                            (3 is the number of channels, all images are RGB)
            * "metadata" : the metadata for each image quotes are the titles for each column
                    0. "image/height"
                    1. "image/width"
                    2. "image/channels"
                    3. "image/class/label"
                    4. "image/object/bbox/xmin"
                    5. "image/object/bbox/xmax"
                    6. "image/object/bbox/ymin"
                    7. "image/object/bbox/ymax"
                    8. "image/object/bbox/label"
            * "file_info" : string information related to each image
                    0. "image/format"
                    1. "image/filename"
                    2. "image/class/synset"
                    3. "image/class/text"


    The dataset was created using the preprocessed data from the script:
        https://github.com/tensorflow/models/blob/master/research/inception/inception/data/download_and_preprocess_imagenet.sh

    """
    import h5py
    import tensorflow as tf

    """
    labels:
        image/encoded: string containing JPEG encoded image in RGB colorspace
        image/height: integer, image height in pixels
        image/width: integer, image width in pixels
        image/colorspace: string, specifying the colorspace, always 'RGB'
        image/channels: integer, specifying the number of channels, always 3
        image/format: string, specifying the format, always 'JPEG'
        image/filename: string containing the basename of the image file
                e.g. 'n01440764_10026.JPEG' or 'ILSVRC2012_val_00000293.JPEG'
        image/class/label: integer specifying the index in a classification layer.
                The label ranges from [1, 1000] where 0 is not used.
        image/class/synset: string specifying the unique ID of the label, e.g. 'n01440764'
        image/class/text: string specifying the human-readable version of the label
                e.g. 'red fox, Vulpes vulpes'
        image/object/bbox/xmin: list of integers specifying the 0+ human annotated bounding boxes
        image/object/bbox/xmax: list of integers specifying the 0+ human annotated bounding boxes
        image/object/bbox/ymin: list of integers specifying the 0+ human annotated bounding boxes
        image/object/bbox/ymax: list of integers specifying the 0+ human annotated bounding boxes
        image/object/bbox/label: integer specifying the index in a classification
                layer. The label ranges from [1, 1000] where 0 is not used. Note this is
                always identical to the image label."""
    # get the number of files from the contents of the folder
    train_names = [folder_name + f for f in os.listdir(folder_name) if f.startswith("train")].sort()
    val_names = [folder_name + f for f in os.listdir(folder_name) if f.startswith("val")].sort()
    num_train = len(train_names)
    num_val = len(val_names)

    def _find_output_name_and_stsp(num_names):
        start = 0
        stop = num_names + 1
        output_name_lcl = output_folder
        output_name_lcl += "imagenet_merged.h5"
        return start, stop, output_name_lcl

    train_start, train_stop, output_name_lcl_train = _find_output_name_and_stsp(num_train)
    val_start, val_stop, output_name_lcl_val = _find_output_name_and_stsp(num_val)
    output_name_lcl_val = f"{output_name_lcl_val[:-3]}_validation.h5"

    # create the output files
    train_lcl_file = h5py.File(output_name_lcl_train, "w")
    dt = h5py.string_dtype(encoding="ascii")
    train_lcl_file.create_dataset("images", (2502,), chunks=(1251,), maxshape=(None,), dtype=dt)
    train_lcl_file.create_dataset("metadata", (2502, 9), chunks=(1251, 9), maxshape=(None, 9))
    train_lcl_file.create_dataset(
        "file_info", (2502, 4), chunks=(1251, 4), maxshape=(None, 4), dtype="S10"
    )

    val_lcl_file = h5py.File(output_name_lcl_val, "w")
    val_lcl_file.create_dataset("images", (50000,), chunks=True, maxshape=(None,), dtype=dt)
    val_lcl_file.create_dataset("metadata", (50000, 9), chunks=True, maxshape=(None, 9))
    val_lcl_file.create_dataset(
        "file_info", (50000, 4), chunks=True, maxshape=(None, 4), dtype="S10"
    )

    def __single_file_load(src):
        # load a file and read it to a numpy array
        dataset = tf.data.TFRecordDataset(filenames=[src])
        imgs = []
        img_meta = [[] for _ in range(9)]
        file_arr = [[] for _ in range(4)]
        for raw_example in iter(dataset):
            parsed = tf.train.Example.FromString(raw_example.numpy())
            img_str = parsed.features.feature["image/encoded"].bytes_list.value[0]
            img = tf.image.decode_jpeg(img_str, channels=3).numpy()
            string_repr = base64.binascii.b2a_base64(img).decode("ascii")
            imgs.append(string_repr)
            # to decode: np.frombuffer(base64.binascii.a2b_base64(string_repr.encode('ascii')))
            img_meta[0].append(
                tf.cast(
                    parsed.features.feature["image/height"].int64_list.value[0], tf.float32
                ).numpy()
            )
            img_meta[1].append(
                tf.cast(
                    parsed.features.feature["image/width"].int64_list.value[0], tf.float32
                ).numpy()
            )
            img_meta[2].append(
                tf.cast(
                    parsed.features.feature["image/channels"].int64_list.value[0], tf.float32
                ).numpy()
            )
            img_meta[3].append(parsed.features.feature["image/class/label"].int64_list.value[0] - 1)
            try:
                bbxmin = parsed.features.feature["image/object/bbox/xmin"].float_list.value[0]
                bbxmax = parsed.features.feature["image/object/bbox/xmax"].float_list.value[0]
                bbymin = parsed.features.feature["image/object/bbox/ymin"].float_list.value[0]
                bbymax = parsed.features.feature["image/object/bbox/ymax"].float_list.value[0]
                bblabel = parsed.features.feature["image/object/bbox/label"].int64_list.value[0] - 1
            except IndexError:
                bbxmin = 0.0
                bbxmax = img_meta[1][-1]
                bbymin = 0.0
                bbymax = img_meta[0][-1]
                bblabel = -2

            img_meta[4].append(np.float(bbxmin))
            img_meta[5].append(np.float(bbxmax))
            img_meta[6].append(np.float(bbymin))
            img_meta[7].append(np.float(bbymax))
            img_meta[8].append(bblabel)

            file_arr[0].append(parsed.features.feature["image/format"].bytes_list.value[0])
            file_arr[1].append(parsed.features.feature["image/filename"].bytes_list.value[0])
            file_arr[2].append(parsed.features.feature["image/class/synset"].bytes_list.value[0])
            file_arr[3].append(
                np.array(parsed.features.feature["image/class/text"].bytes_list.value[0])
            )
        # need to transpose because of the way that numpy understands nested lists
        img_meta = np.array(img_meta, dtype=np.float64).T
        file_arr = np.array(file_arr).T
        return imgs, img_meta, file_arr

    def __write_datasets(img_outl, img_metal, file_arrl, past_sizel, file):
        file["images"].resize((past_sizel + len(img_outl),))
        file["images"][past_sizel : len(img_outl) + past_sizel] = img_outl
        file["metadata"].resize((past_sizel + img_metal.shape[0], 9))
        file["metadata"][past_sizel : img_metal.shape[0] + past_sizel] = img_metal
        file["file_info"].resize((past_sizel + img_metal.shape[0], 4))
        file["file_info"][past_sizel : img_metal.shape[0] + past_sizel] = file_arrl

    def __load_multiple_files(train_names, train_start, train_stop, file):
        loc_files = train_names[train_start:train_stop]
        img_out, img_meta, file_arr = None, None, None
        past_size, i = 0, 0
        for f in loc_files:  # train
            # print(f)
            # this is where the data is created for
            imgs, img_metaf, file_arrf = __single_file_load(f)
            # create a larger ndarray with the results
            if img_out is not None:
                img_out.extend(imgs)
            else:
                img_out = imgs
            img_meta = np.vstack((img_meta, img_metaf)) if img_meta is not None else img_metaf
            file_arr = np.vstack((file_arr, file_arrf)) if file_arr is not None else file_arrf
            # when 2 files are read, write to the output file
            if i % 2 == 1:
                print(past_size)
                __write_datasets(img_out, img_meta, file_arr, past_size, file)
                past_size += len(img_out)
                img_out, img_meta, file_arr = None, None, None
                del imgs, img_metaf, file_arrf
            i += 1

        if img_out is not None:
            __write_datasets(img_out, img_meta, file_arr, past_size, file)

    __load_multiple_files(train_names, train_start, train_stop, train_lcl_file)
    __load_multiple_files(val_names, val_start, val_stop, val_lcl_file)

    #  add the label names to the datasets
    img_list = [1, 2, 4, 7, 10, 11, 12, 13, 14]
    file_list = [5, 6, 8, 9]
    feature_list = [
        "image/encoded",
        "image/height",
        "image/width",
        "image/colorspace",
        "image/channels",
        "image/format",
        "image/filename",
        "image/class/label",
        "image/class/synset",
        "image/class/text",
        "image/object/bbox/xmin",
        "image/object/bbox/xmax",
        "image/object/bbox/ymin",
        "image/object/bbox/ymax",
        "image/object/bbox/label",
    ]

    train_lcl_file["metadata"].attrs["column_names"] = [feature_list[im] for im in img_list]
    train_lcl_file["file_info"].attrs["column_names"] = [feature_list[im] for im in file_list]
    val_lcl_file["metadata"].attrs["column_names"] = [feature_list[im] for im in img_list]
    val_lcl_file["file_info"].attrs["column_names"] = [feature_list[im] for im in file_list]



"""
Generate matrices for specific tests and functions
"""

from heat import core
from ...core.dndarray import DNDarray
from ...core.communication import Communication
from ...core.devices import Device
from ...core.types import datatype
from ...core.random import randn, rand
from ...core.linalg import qr, matmul
from ...core.manipulations import diag, sort
from ...core.exponential import log
from typing import Type, Union, Tuple, Callable

__all__ = ["parter", "random_known_singularvalues", "random_known_rank"]


def parter(
    n: int,
    split: Union[None, int] = None,
    device: Union[None, str, Device] = None,
    comm: Union[None, Communication] = None,
    dtype: Type[datatype] = core.float32,
) -> DNDarray:
    """
    Generates the Parter matrix, a Toeplitz matrix that has the interesting property of having its singular values cluster at
    pi. The matrix has been named so by Cleve Moler in recognition of Seymour Parter's proof of this fact.

    Parameters
    ----------
    n : int
        size of the resulting square matrix
    split: None or int, optional
        The axis along which the array content is split and distributed in memory.
    device: None or str or Device, optional
        Specifies the device the tensor shall be allocated on, defaults globally set default device.
    comm: None or Communication, optional
        Handle to the nodes holding distributed tensor chunks.
    dtype: Type[datatype], optional
        The desired data-type for the array, defaults to ht.float64.

    References
    ----------
    [1] https://blogs.mathworks.com/cleve/2019/06/24/bohemian-matrices-in-the-matlab-gallery/

    [2] https://blogs.mathworks.com/cleve/2014/02/03/surprising-svd-square-waves-and-pi/

    [3] Seymour V. Parter, On the distribution of the singular values of Toeplitz matrices, Linear Algebra and its
    Applications 80, 1986, 115-130, http://www.sciencedirect.com/science/article/pii/0024379586902806
    """
    if split is None:
        a = core.arange(n, dtype=dtype, device=device, comm=comm)
        II = a.expand_dims(0)
        JJ = a.expand_dims(1)
    elif split == 0:
        II = core.arange(n, dtype=dtype, device=device, comm=comm).expand_dims(0)
        JJ = core.arange(n, dtype=dtype, split=split, device=device, comm=comm).expand_dims(1)
    elif split == 1:
        II = core.arange(n, dtype=dtype, split=0, device=device, comm=comm).expand_dims(0)
        JJ = core.arange(n, dtype=dtype, device=device, comm=comm).expand_dims(1)
    else:
        raise ValueError("expected split value to be either {{None,0,1}}, but was {}".format(split))

    return 1.0 / (II - JJ + 0.5)


def random_orthogonal(
    m: int,
    n: int,
    split: Union[None, int] = None,
    device: Union[None, str, Device] = None,
    comm: Union[None, Communication] = None,
    dtype: Type[datatype] = core.float32,
) -> DNDarray:
    """Auxiliary routine: creates a random mxn matrix with orthonormal columns
    Caveat: this is done by QR of mxn matrices with i.i.d. normal entries, so this does not produce the uniform distribution on the orthogonal matrices...
    """
    if n > m:
        raise RuntimeError("No orthogonal matrix of shape %d x %d possible." % (m, n))

    # TODO: if QR does not make problems anymore, replace split=None by split=split
    U = randn(m, n, split=None, dtype=dtype, comm=comm, device=device)
    Q, _ = qr(U)

    return Q[:, :n].resplit_(split)


def random_known_singularvalues(
    m: int,
    n: int,
    singular_values: DNDarray,
    split: Union[None, int] = None,
    device: Union[None, str, Device] = None,
    comm: Union[None, Communication] = None,
    dtype: Type[datatype] = core.float32,
) -> Tuple[DNDarray, Tuple[DNDarray]]:
    """
    Creates an m x n matrix with singular values given by the entries of the input array singular_values.
    Caveat: if the entries of `singular_values` are not sorted, the singular value decomposition of A (returned as second output) is so as well.
    The singular vectors are chosen randomly using :func:`random_orthogonal`.
    """
    if not isinstance(singular_values, DNDarray):
        raise RuntimeError(
            "Argument singular_values needs to be a DNDarray but is {}.".format(
                type(singular_values)
            )
        )
    if not singular_values.ndim == 1:
        raise RuntimeError(
            "Argument singular_values needs to be a 1D array, but dimension is {}.".format(
                singular_values.ndim
            )
        )
    if singular_values.shape[0] > min(m, n):
        raise RuntimeError(
            f"Number of given singular values must not exceed matrix dimensions. Got {singular_values.shape[0]} singular values for matrix size ({m}, {n})."
        )

    r = singular_values.shape[0]
    U = random_orthogonal(m, r, split=split, device=device, comm=comm, dtype=dtype)
    V = random_orthogonal(n, r, split=split, device=device, comm=comm, dtype=dtype)

    A = matmul(U, matmul(diag(singular_values), V.T))

    return A.resplit_(split), (U, singular_values, V)


def random_known_rank(
    m: int,
    n: int,
    r: int,
    quantile_function: Callable = lambda x: -log(x),
    split: Union[None, int] = None,
    device: Union[None, str, Device] = None,
    comm: Union[None, Communication] = None,
    dtype: Type[datatype] = core.float32,
) -> Tuple[DNDarray, Tuple[DNDarray]]:
    """
    Creates a random m x n matrix with rank r.
    This routine uses :func:`random_known_singularvalues` with r singular values randomly chosen
    w.r.t. the distribution with quantile function given by the input quantile_function. Default yields exponential distibution with parameter lambda=1.
    Unlike in :func:`random_known_singularvalues`, here the singular values of the output are sorted in descending order.
    """
    if r > min(m, n):
        raise RuntimeError("rank must not exceed matrix dimensions.")

    singular_values = rand(r, dtype=dtype, comm=comm, device=device)
    singular_values = sort(quantile_function(singular_values), descending=True)[0]

    return random_known_singularvalues(
        m, n, singular_values, split=split, device=device, comm=comm, dtype=dtype
    )



"""Provides DCSR_matrix, a distributed compressed sparse row matrix"""
from __future__ import annotations

import torch
from mpi4py import MPI
from typing import Union, Tuple, TypeVar

from ..core.devices import Device
from ..core.dndarray import DNDarray
from ..core.factories import array
from ..core.types import datatype, canonical_heat_type

__all__ = ["DCSR_matrix"]

Communication = TypeVar("Communication")


class DCSR_matrix:
    """
    Distributed Compressed Sparse Row Matrix. It is composed of
    PyTorch sparse_csr_tensors local to each process.

    Parameters
    ----------
    array : torch.Tensor (layout ==> torch.sparse_csr)
        Local sparse array
    gnnz: int
        Total number of non-zero elements across all processes
    gshape : Tuple[int,...]
        The global shape of the array
    dtype : datatype
        The datatype of the array
    split : int or None
        If split is not None, it denotes the axis on which the array is divided between processes.
        DCSR_matrix only supports distribution along axis 0.
    device : Device
        The device on which the local arrays are using (cpu or gpu)
    comm : Communication
        The communications object for sending and receiving data
    balanced: bool or None
        Describes whether the data are evenly distributed across processes.
    """

    def __init__(
        self,
        array: torch.Tensor,
        gnnz: int,
        gshape: Tuple[int, ...],
        dtype: datatype,
        split: Union[int, None],
        device: Device,
        comm: Communication,
        balanced: bool,
    ):
        self.__array = array
        self.__gnnz = gnnz
        self.__gshape = gshape
        self.__dtype = dtype
        self.__split = split
        self.__device = device
        self.__comm = comm
        self.__balanced = balanced

    def global_indptr(self) -> DNDarray:
        """
        Global indptr of the ``DCSR_matrix`` as a ``DNDarray``
        """
        if self.split is None:
            raise ValueError("This method works only for distributed matrices")

        # Need to know the number of non-zero elements
        # in the processes with lesser rank
        all_nnz = torch.zeros(self.comm.size + 1, device=self.device.torch_device)

        # Each process must drop their nnz in index = rank + 1
        all_nnz[self.comm.rank + 1] = self.lnnz
        self.comm.Allreduce(MPI.IN_PLACE, all_nnz, MPI.SUM)

        # Build prefix array out of all the nnz
        all_nnz = torch.cumsum(all_nnz, dim=0)

        global_indptr = self.lindptr + int(all_nnz[self.comm.rank])

        # Remove the (n+1) the element from all the processes except last
        if self.comm.rank != self.comm.size - 1:
            global_indptr = global_indptr[:-1]

        # NOTE: indptr might be unbalanced in distribution but should not be self balanced
        return array(
            global_indptr,
            dtype=self.lindptr.dtype,
            device=self.device,
            comm=self.comm,
            is_split=self.split,
        )

    @property
    def balanced(self) -> bool:
        """
        Boolean value indicating if the DCSR_matrix is balanced between the MPI processes
        """
        return self.__balanced

    @property
    def comm(self) -> Communication:
        """
        The :class:`~heat.core.communication.Communication` of the ``DCSR_matrix``
        """
        return self.__comm

    @property
    def device(self) -> Device:
        """
        The :class:`~heat.core.devices.Device` of the ``DCSR_matrix``
        """
        return self.__device

    @property
    def larray(self) -> torch.Tensor:
        """
        Local data of the ``DCSR_matrix``
        """
        return self.__array

    @property
    def data(self) -> torch.Tensor:
        """
        Global data of the ``DCSR_matrix``
        """
        if self.split is None:
            return self.ldata

        data_buffer = torch.zeros(
            size=(self.gnnz,), dtype=self.dtype.torch_type(), device=self.device.torch_device
        )
        counts, displs = self.counts_displs_nnz()
        self.comm.Allgatherv(self.ldata, (data_buffer, counts, displs))
        return data_buffer

    @property
    def gdata(self) -> torch.Tensor:
        """
        Global data of the ``DCSR_matrix``
        """
        return self.data

    @property
    def ldata(self) -> torch.Tensor:
        """
        Local data of the ``DCSR_matrix``
        """
        return self.__array.values()

    @property
    def indptr(self) -> torch.Tensor:
        """
        Global indptr of the ``DCSR_matrix``
        """
        if self.split is None:
            return self.lindptr

        return self.global_indptr().resplit(axis=None).larray

    @property
    def gindptr(self) -> torch.Tensor:
        """
        Global indptr of the ``DCSR_matrix``
        """
        return self.indptr

    @property
    def lindptr(self) -> torch.Tensor:
        """
        Local indptr of the ``DCSR_matrix``
        """
        return self.__array.crow_indices()

    @property
    def indices(self) -> torch.Tensor:
        """
        Global indices of the ``DCSR_matrix``
        """
        if self.split is None:
            return self.lindices

        indices_buffer = torch.zeros(
            size=(self.gnnz,), dtype=self.lindices.dtype, device=self.device.torch_device
        )
        counts, displs = self.counts_displs_nnz()
        self.comm.Allgatherv(self.lindices, (indices_buffer, counts, displs))
        return indices_buffer

    @property
    def gindices(self) -> torch.Tensor:
        """
        Global indices of the ``DCSR_matrix``
        """
        return self.indices

    @property
    def lindices(self) -> torch.Tensor:
        """
        Local indices of the ``DCSR_matrix``
        """
        return self.__array.col_indices()

    @property
    def ndim(self) -> int:
        """
        Number of dimensions of the ``DCSR_matrix``
        """
        return len(self.__gshape)

    @property
    def nnz(self) -> int:
        """
        Total number of non-zero elements of the ``DCSR_matrix``
        """
        return self.__gnnz

    @property
    def gnnz(self) -> int:
        """
        Total number of non-zero elements of the ``DCSR_matrix``
        """
        return self.nnz

    @property
    def lnnz(self) -> int:
        """
        Number of non-zero elements on the local process of the ``DCSR_matrix``
        """
        return self.__array._nnz()

    @property
    def shape(self) -> Tuple[int, ...]:
        """
        Global shape of the ``DCSR_matrix``
        """
        return self.__gshape

    @property
    def gshape(self) -> Tuple[int, ...]:
        """
        Global shape of the ``DCSR_matrix``
        """
        return self.shape

    @property
    def lshape(self) -> Tuple[int, ...]:
        """
        Local shape of the ``DCSR_matrix``
        """
        return tuple(self.__array.size())

    @property
    def dtype(self):
        """
        The :class:`~heat.core.types.datatype` of the ``DCSR_matrix``
        """
        return self.__dtype

    @property
    def split(self) -> int:
        """
        Returns the axis on which the ``DCSR_matrix`` is split
        """
        return self.__split

    def is_distributed(self) -> bool:
        """
        Determines whether the data of this ``DCSR_matrix`` is distributed across multiple processes.
        """
        return self.split is not None and self.comm.is_distributed()

    def counts_displs_nnz(self) -> Tuple[Tuple[int], Tuple[int]]:
        """
        Returns actual counts (number of non-zero items per process) and displacements (offsets) of the DCSR_matrix.
        Does not assume load balance.
        """
        if self.split is not None:
            counts = torch.zeros(self.comm.size)
            counts[self.comm.rank] = self.lnnz
            self.comm.Allreduce(MPI.IN_PLACE, counts, MPI.SUM)
            displs = [0] + torch.cumsum(counts, dim=0)[:-1].tolist()
            return tuple(counts.tolist()), tuple(displs)
        else:
            raise ValueError(
                "Non-distributed DCSR_matrix. Cannot calculate counts and displacements."
            )

    def astype(self, dtype, copy=True) -> DCSR_matrix:
        """
        Returns a casted version of this matrix.
        Casted matrix is a new matrix of the same shape but with given type of this matrix. If copy is ``True``, the
        same matrix is returned instead.

        Parameters
        ----------
        dtype : datatype
            HeAT type to which the matrix is cast
        copy : bool, optional
            By default the operation returns a copy of this matrix. If copy is set to ``False`` the cast is performed
            in-place and this matrix is returned
        """
        dtype = canonical_heat_type(dtype)
        casted_matrix = self.__array.type(dtype.torch_type())
        if copy:
            return DCSR_matrix(
                casted_matrix,
                self.gnnz,
                self.gshape,
                dtype,
                self.split,
                self.device,
                self.comm,
                self.balanced,
            )

        self.__array = casted_matrix
        self.__dtype = dtype

        return self

    def __repr__(self) -> str:
        """
        Computes a printable representation of the passed DCSR_matrix.
        """
        print_string = (
            f"(indptr: {self.indptr}, indices: {self.indices}, data: {self.data}, "
            f"dtype=ht.{self.dtype.__name__}, device={self.device}, split={self.split})"
        )

        # Check has to happen after generating string because
        # generation of string invokes functions that require
        # participation from all processes
        if self.comm.rank != 0:
            return ""
        return print_string



"""Generalized operations for DNDarray"""

import builtins
import numpy as np
import torch
import warnings

from .communication import MPI, MPI_WORLD
from . import factories
from . import stride_tricks
from . import sanitation
from . import statistics
from .dndarray import DNDarray
from . import types

from typing import Callable, Optional, Type, Union, Dict

__all__ = []
__BOOLEAN_OPS = [MPI.LAND, MPI.LOR, MPI.BAND, MPI.BOR]


def __binary_op(
    operation: Callable,
    t1: Union[DNDarray, int, float],
    t2: Union[DNDarray, int, float],
    out: Optional[DNDarray] = None,
    where: Optional[DNDarray] = None,
    fn_kwargs: Optional[Dict] = {},
) -> DNDarray:
    """
    Generic wrapper for element-wise binary operations of two operands (either can be tensor or scalar).
    Takes the operation function and the two operands involved in the operation as arguments.

    Parameters
    ----------
    operation : function
        The operation to be performed. Function that performs operation elements-wise on the involved tensors,
        e.g. add values from other to self
    t1: DNDarray or scalar
        The first operand involved in the operation.
    t2: DNDarray or scalar
        The second operand involved in the operation.
    out: DNDarray, optional
        Output buffer in which the result is placed. If not provided, a freshly allocated array is returned.
    where: DNDarray, optional
        Condition to broadcast over the inputs. At locations where the condition is True, the `out` array
        will be set to the result of the operation. Elsewhere, the `out` array will retain its original
        value. If an uninitialized `out` array is created via the default `out=None`, locations within
        it where the condition is False will remain uninitialized. If distributed, the split axis (after
        broadcasting if required) must match that of the `out` array.
    fn_kwargs: Dict, optional
        keyword arguments used for the given operation
        Default: {} (empty dictionary)

    Returns
    -------
    result: ht.DNDarray
        A DNDarray containing the results of element-wise operation.

    Warning
    -------
    If both operands are distributed, they must be distributed along the same dimension, i.e. `t1.split = t2.split`.

    MPI communication is necessary when both operands are distributed along the same dimension, but the distribution maps do not match. E.g.:
    ```
    a =  ht.ones(10000, split=0)
    b = ht.zeros(10000, split=0)
    c = a[:-1] + b[1:]
    ```
    In such cases, one of the operands is redistributed OUT-OF-PLACE to match the distribution map of the other operand.
    The operand determining the resulting distribution is chosen as follows:
    1) split is preferred to no split
    2) no (shape)-broadcasting in the split dimension if not necessary
    3) t1 is preferred to t2
    """
    # Check inputs
    if not np.isscalar(t1) and not isinstance(t1, DNDarray):
        raise TypeError(
            f"Only DNDarrays and numeric scalars are supported, but input was {type(t1)}"
        )
    if not np.isscalar(t2) and not isinstance(t2, DNDarray):
        raise TypeError(
            f"Only DNDarrays and numeric scalars are supported, but input was {type(t2)}"
        )
    promoted_type = types.result_type(t1, t2).torch_type()

    # Make inputs Dndarrays
    if np.isscalar(t1) and np.isscalar(t2):
        try:
            t1 = factories.array(t1)
            t2 = factories.array(t2)
        except (ValueError, TypeError):
            raise TypeError(f"Data type not supported, inputs were {type(t1)} and {type(t2)}")
    elif np.isscalar(t1) and isinstance(t2, DNDarray):
        try:
            t1 = factories.array(t1, device=t2.device, comm=t2.comm)
        except (ValueError, TypeError):
            raise TypeError(f"Data type not supported, input was {type(t1)}")
    elif isinstance(t1, DNDarray) and np.isscalar(t2):
        try:
            t2 = factories.array(t2, device=t1.device, comm=t1.comm)
        except (ValueError, TypeError):
            raise TypeError(f"Data type not supported, input was {type(t2)}")

    # Make inputs have the same dimensionality
    output_shape = stride_tricks.broadcast_shape(t1.shape, t2.shape)
    if where is not None:
        output_shape = stride_tricks.broadcast_shape(where.shape, output_shape)
        while len(where.shape) < len(output_shape):
            where = where.expand_dims(axis=0)
    # Broadcasting allows additional empty dimensions on the left side
    # TODO simplify this once newaxis-indexing is supported to get rid of the loops
    while len(t1.shape) < len(output_shape):
        t1 = t1.expand_dims(axis=0)
    while len(t2.shape) < len(output_shape):
        t2 = t2.expand_dims(axis=0)
    # t1 = t1[tuple([None] * (len(output_shape) - t1.ndim))]
    # t2 = t2[tuple([None] * (len(output_shape) - t2.ndim))]
    # print(t1.lshape, t2.lshape)

    def __get_out_params(target, other=None, map=None):
        """
        Getter for the output parameters of a binary operation with target distribution.
        If `other` is provided, its distribution will be matched to `target` or, if provided,
        redistributed according to `map`.

        Parameters
        ----------
        target : DNDarray
            DNDarray determining the parameters
        other : DNDarray
            DNDarray to be adapted
        map : Tensor
            lshape_map `other` should be matched to. Defaults to `target.lshape_map`

        Returns
        -------
        Tuple
            split, device, comm, balanced, [other]
        """
        if other is not None:
            if out is None:
                other = sanitation.sanitize_distribution(other, target=target, diff_map=map)
            return target.split, target.device, target.comm, target.balanced, other
        return target.split, target.device, target.comm, target.balanced

    if t1.split is not None and t1.shape[t1.split] == output_shape[t1.split]:  # t1 is "dominant"
        output_split, output_device, output_comm, output_balanced, t2 = __get_out_params(t1, t2)
    elif t2.split is not None and t2.shape[t2.split] == output_shape[t2.split]:  # t2 is "dominant"
        output_split, output_device, output_comm, output_balanced, t1 = __get_out_params(t2, t1)
    elif t1.split is not None:
        # t1 is split but broadcast -> only on one rank; manipulate lshape_map s.t. this rank has 'full' data
        lmap = t1.lshape_map
        idx = lmap[:, t1.split].nonzero(as_tuple=True)[0]
        lmap[idx.item(), t1.split] = output_shape[t1.split]
        output_split, output_device, output_comm, output_balanced, t2 = __get_out_params(
            t1, t2, map=lmap
        )
    elif t2.split is not None:
        # t2 is split but broadcast -> only on one rank; manipulate lshape_map s.t. this rank has 'full' data
        lmap = t2.lshape_map
        idx = lmap[:, t2.split].nonzero(as_tuple=True)[0]
        lmap[idx.item(), t2.split] = output_shape[t2.split]
        output_split, output_device, output_comm, output_balanced, t1 = __get_out_params(
            t2, other=t1, map=lmap
        )
    else:  # both are not split
        output_split, output_device, output_comm, output_balanced = __get_out_params(t1)

    if out is not None:
        sanitation.sanitize_out(out, output_shape, output_split, output_device, output_comm)
        t1, t2 = sanitation.sanitize_distribution(t1, t2, target=out)

    result = operation(t1.larray.to(promoted_type), t2.larray.to(promoted_type), **fn_kwargs)

    if out is None and where is None:
        return DNDarray(
            result,
            output_shape,
            types.heat_type_of(result),
            output_split,
            device=output_device,
            comm=output_comm,
            balanced=output_balanced,
        )

    if where is not None:
        if out is None:
            out = factories.empty(
                output_shape,
                dtype=promoted_type,
                split=output_split,
                device=output_device,
                comm=output_comm,
            )
        if where.split != out.split:
            where = sanitation.sanitize_distribution(where, target=out)
        result = torch.where(where.larray, result, out.larray)

    out.larray.copy_(result)
    return out


def __cum_op(
    x: DNDarray,
    partial_op: Callable,
    exscan_op: Callable,
    final_op: Callable,
    neutral: Union[int, float],
    axis: Union[int, float],
    dtype: Union[str, Type[types.datatype]],
    out: Optional[DNDarray] = None,
) -> DNDarray:
    """
    Generic wrapper for cumulative operations. Performs a three-stage cumulative operation. First, a partial
    cumulative operation is performed node-local that is combined into a global cumulative result via an MPI_Op and a final local
    reduction add or mul operation.

    Parameters
    ----------
    x : DNDarray
        The heat DNDarray on which to perform the cumulative operation
    partial_op: function
        The function performing a partial cumulative operation on the process-local data portion, e.g. :func:`cumsum() <heat.arithmetics.cumsum>`.
    exscan_op: mpi4py.MPI.Op
        The MPI operator for performing the exscan based on the results returned by the partial_op function.
    final_op: function
        The local operation for the final result, e.g. :func:`add() <heat.arithmetics.add>` for :func:`cumsum() <heat.arithmetics.cumsum>`.
    neutral: scalar
        Neutral element for the cumulative operation, i.e. an element that does not change the reductions operations
        result.
    axis: int
        The axis direction of the cumulative operation
    dtype: datatype
        The type of the result tensor.
    out: DNDarray, optional
        The explicitly returned output tensor.

    Raises
    ------
    TypeError
        If the input or optional output parameter are not of type ht.DNDarray
    ValueError
        If the shape of the optional output parameters does not match the shape of the input
    NotImplementedError
        Numpy's behaviour of axis is None is not supported as of now
    RuntimeError
        If the split or device parameters do not match the parameters of the input
    """
    # perform sanitation
    sanitation.sanitize_in(x)

    if axis is None:
        raise NotImplementedError("axis = None is not supported")
    axis = stride_tricks.sanitize_axis(x.shape, axis)

    if dtype is not None:
        dtype = types.canonical_heat_type(dtype)

    if out is not None:
        sanitation.sanitize_out(out, x.shape, x.split, x.device)
        dtype = out.dtype

    cumop = partial_op(
        x.larray,
        axis,
        out=None if out is None else out.larray,
        dtype=None if dtype is None else dtype.torch_type(),
    )

    if x.split is not None and axis == x.split:
        indices = torch.tensor([cumop.shape[axis] - 1], device=cumop.device)
        send = (
            torch.index_select(cumop, axis, indices)
            if indices[0] >= 0
            else torch.full(
                cumop.shape[:axis] + torch.Size([1]) + cumop.shape[axis + 1 :],
                neutral,
                dtype=cumop.dtype,
                device=cumop.device,
            )
        )
        recv = torch.full(
            cumop.shape[:axis] + torch.Size([1]) + cumop.shape[axis + 1 :],
            neutral,
            dtype=cumop.dtype,
            device=cumop.device,
        )

        x.comm.Exscan(send, recv, exscan_op)
        final_op(cumop, recv, out=cumop)

    if out is not None:
        return out

    return factories.array(
        cumop,
        dtype=x.dtype if dtype is None else dtype,
        is_split=x.split,
        device=x.device,
        comm=x.comm,
    )


def __local_op(
    operation: Callable,
    x: DNDarray,
    out: Optional[DNDarray] = None,
    no_cast: Optional[bool] = False,
    **kwargs,
) -> DNDarray:
    """
    Generic wrapper for local operations, which do not require communication. Accepts the actual operation function as
    argument and takes only care of buffer allocation/writing. This function is intended to work on an element-wise bases

    Parameters
    ----------
    operation : function
        A function implementing the element-wise local operation, e.g. torch.sqrt
    x : DNDarray
        The value for which to compute 'operation'.
    no_cast : bool
        Flag to avoid casting to floats
    out : DNDarray, optional
        A location in which to store the results. If provided, it must have a broadcastable shape. If not provided or
        set to None, a fresh tensor is allocated.

    Warning
    -------
    The gshape of the result DNDarray will be the same as that of x

    Raises
    -------
    TypeError
        If the input is not a tensor or the output is not a tensor or None.
    """
    # perform sanitation
    sanitation.sanitize_in(x)
    if out is not None and not isinstance(out, DNDarray):
        raise TypeError(f"expected out to be None or an ht.DNDarray, but was {type(out)}")

    # infer the output type of the tensor
    # we need floating point numbers here, due to PyTorch only providing sqrt() implementation for float32/64
    if not no_cast:
        promoted_type = types.promote_types(x.dtype, types.float32)
        torch_type = promoted_type.torch_type()
    else:
        torch_type = x.larray.dtype

    # no defined output tensor, return a freshly created one
    if out is None:
        result = operation(x.larray.type(torch_type), **kwargs)
        return DNDarray(
            result,
            x.gshape,
            types.canonical_heat_type(result.dtype),
            x.split,
            x.device,
            x.comm,
            x.balanced,
        )

    # output buffer writing requires a bit more work
    # we need to determine whether the operands are broadcastable and the multiple of the broadcasting
    # reason: manually repetition for each dimension as PyTorch does not conform to numpy's broadcast semantic
    # PyTorch always recreates the input shape and ignores broadcasting for too large buffers
    broadcast_shape = stride_tricks.broadcast_shape(x.lshape, out.lshape)
    padded_shape = (1,) * (len(broadcast_shape) - len(x.lshape)) + x.lshape
    multiples = [(int(a / b) if b > 0 else 0) for a, b in zip(broadcast_shape, padded_shape)]
    needs_repetition = builtins.any(multiple > 1 for multiple in multiples)

    # do an inplace operation into a provided buffer
    casted = x.larray.type(torch_type)
    operation(casted.repeat(multiples) if needs_repetition else casted, out=out.larray, **kwargs)

    return out


def __reduce_op(
    x: DNDarray,
    partial_op: Callable,
    reduction_op: Callable,
    neutral: Optional[Union[int, float]] = None,
    **kwargs,
) -> DNDarray:
    """
    Generic wrapper for reduction operations, e.g. :func:`sum() <heat.arithmetics.sum>`, :func:`prod() <heat.arithmetics.prod>`
    etc. Performs a two-stage reduction. First, a partial reduction is performed node-local that is combined into a
    global reduction result via an MPI_Op.

    Parameters
    ----------
    x : DNDarray
        The DNDarray on which to perform the reduction operation
    partial_op: function
        The function performing a partial reduction on the process-local data portion, e.g. sum() for implementing a
        distributed mean() operation.
    reduction_op: mpi4py.MPI.Op
        The MPI operator for performing the full reduction based on the results returned by the partial_op function.
    neutral: scalar
        Neutral element, i.e. an element that does not change the result of the reduction operation. Needed for
        those cases where 'x.gshape[x.split] < x.comm.rank', that is, the shape of the distributed tensor is such
        that one or more processes will be left without data.

    Raises
    ------
    TypeError
        If the input or optional output parameter are not of type ht.DNDarray
    ValueError
        If the shape of the optional output parameters does not match the shape of the reduced result
    """
    # perform sanitation
    sanitation.sanitize_in(x)

    # no further checking needed, sanitize axis will raise the proper exceptions
    axis = stride_tricks.sanitize_axis(x.shape, kwargs.get("axis"))
    if isinstance(axis, int):
        axis = (axis,)
    keepdims = kwargs.get("keepdims")
    out = kwargs.get("out")
    split = x.split
    balanced = x.balanced

    # if local tensor is empty, replace it with the identity element
    if x.is_distributed() and 0 in x.lshape and (axis is None or split in axis):
        if neutral is None:
            neutral = float("nan")
        neutral_shape = x.gshape[:split] + (1,) + x.gshape[split + 1 :]
        partial = torch.full(
            neutral_shape,
            fill_value=neutral,
            dtype=x.dtype.torch_type(),
            device=x.device.torch_device,
        )
    else:
        partial = x.larray

    # apply the partial reduction operation to the local tensor
    if axis is None:
        partial = partial_op(partial).reshape(-1)
        output_shape = (1,)
        balanced = True
    else:
        output_shape = x.gshape
        for dim in axis:
            if not (
                partial.shape.numel() == 0 and partial_op.__name__ in ("local_max", "local_min")
            ):  # no neutral element for max/min
                partial = partial_op(partial, dim=dim, keepdim=True)
            output_shape = output_shape[:dim] + (1,) + output_shape[dim + 1 :]
        if not keepdims and len(partial.shape) != 1:
            gshape_losedim = tuple(x.gshape[dim] for dim in range(len(x.gshape)) if dim not in axis)
            lshape_losedim = tuple(x.lshape[dim] for dim in range(len(x.lshape)) if dim not in axis)
            output_shape = gshape_losedim
            # Take care of special cases argmin and argmax: keep partial.shape[0]
            if 0 in axis and partial.shape[0] != 1:
                lshape_losedim = (partial.shape[0],) + lshape_losedim
            if 0 not in axis and partial.shape[0] != x.lshape[0]:
                lshape_losedim = (partial.shape[0],) + lshape_losedim[1:]
            if len(lshape_losedim) > 0:
                partial = partial.reshape(lshape_losedim)
    # perform a reduction operation in case the tensor is distributed across the reduction axis
    if x.split is not None:
        if axis is None or (x.split in axis):
            split = None
            balanced = True
            if x.comm.is_distributed():
                x.comm.Allreduce(MPI.IN_PLACE, partial, reduction_op)
        elif axis is not None and not keepdims:
            down_dims = len(tuple(dim for dim in axis if dim < x.split))
            split -= down_dims
            balanced = x.balanced

    ARG_OPS = [statistics.MPI_ARGMAX, statistics.MPI_ARGMIN]
    arg_op = False
    if reduction_op in ARG_OPS:
        arg_op = True
        partial = partial.chunk(2)[-1].type(torch.int64)
        if partial.ndim > 1:
            partial = partial.squeeze(dim=0)

    # if reduction_op is a Boolean operation, then resulting tensor is bool
    tensor_type = bool if reduction_op in __BOOLEAN_OPS else partial.dtype

    if out is not None:
        # sanitize out
        sanitation.sanitize_out(out, output_shape, split, x.device)
        if arg_op and out.dtype != types.canonical_heat_type(partial.dtype):
            raise TypeError(
                f"Data type mismatch: out.dtype should be {types.canonical_heat_type(partial.dtype)}, is {out.dtype}"
            )
        out._DNDarray__array = partial
        return out

    return DNDarray(
        partial,
        output_shape,
        types.canonical_heat_type(tensor_type),
        split=split,
        device=x.device,
        comm=x.comm,
        balanced=balanced,
    )



"""Arithmetic functions for Dcsr_matrices"""
from __future__ import annotations

import torch

from .dcsr_matrix import DCSR_matrix

from . import _operations

__all__ = [
    "add",
    "mul",
]


def add(t1: DCSR_matrix, t2: DCSR_matrix) -> DCSR_matrix:
    """
    Element-wise addition of values from two operands, commutative.
    Takes the first and second operand (scalar or :class:`~heat.sparse.DCSR_matrix`) whose elements are to be added
    as argument and returns a ``DCSR_matrix`` containing the results of element-wise addition of ``t1`` and ``t2``.

    Parameters
    ----------
    t1: DCSR_matrix
        The first operand involved in the addition
    t2: DCSR_matrix
        The second operand involved in the addition

    Examples
    --------
    >>> heat_sparse_csr
    (indptr: tensor([0, 2, 3]), indices: tensor([0, 2, 2]), data: tensor([1., 2., 3.]), dtype=ht.float32, device=cpu:0, split=0)
    >>> heat_sparse_csr.todense()
    DNDarray([[1., 0., 2.],
              [0., 0., 3.]], dtype=ht.float32, device=cpu:0, split=0)
    >>> sum_sparse = heat_sparse_csr + heat_sparse_csr
        (or)
    >>> sum_sparse = ht.sparse.sparse_add(heat_sparse_csr, heat_sparse_csr)
    >>> sum_sparse
    (indptr: tensor([0, 2, 3], dtype=torch.int32), indices: tensor([0, 2, 2], dtype=torch.int32), data: tensor([2., 4., 6.]), dtype=ht.float32, device=cpu:0, split=0)
    >>> sum_sparse.todense()
    DNDarray([[2., 0., 4.],
              [0., 0., 6.]], dtype=ht.float32, device=cpu:0, split=0)
    """
    return _operations.__binary_op_csr(torch.add, t1, t2)


DCSR_matrix.__add__ = lambda self, other: add(self, other)
DCSR_matrix.__add__.__doc__ = add.__doc__
DCSR_matrix.__radd__ = lambda self, other: add(self, other)
DCSR_matrix.__radd__.__doc__ = add.__doc__


def mul(t1: DCSR_matrix, t2: DCSR_matrix) -> DCSR_matrix:
    """
    Element-wise multiplication (NOT matrix multiplication) of values from two operands, commutative.
    Takes the first and second operand (scalar or :class:`~heat.sparse.DCSR_matrix`) whose elements are to be
    multiplied as argument.

    Parameters
    ----------
    t1: DCSR_matrix
        The first operand involved in the multiplication
    t2: DCSR_matrix
        The second operand involved in the multiplication

    Examples
    --------
    >>> heat_sparse_csr
    (indptr: tensor([0, 2, 3]), indices: tensor([0, 2, 2]), data: tensor([1., 2., 3.]), dtype=ht.float32, device=cpu:0, split=0)
    >>> heat_sparse_csr.todense()
    DNDarray([[1., 0., 2.],
              [0., 0., 3.]], dtype=ht.float32, device=cpu:0, split=0)
    >>> pdt_sparse = heat_sparse_csr * heat_sparse_csr
        (or)
    >>> pdt_sparse = ht.sparse.sparse_mul(heat_sparse_csr, heat_sparse_csr)
    >>> pdt_sparse
    (indptr: tensor([0, 2, 3]), indices: tensor([0, 2, 2]), data: tensor([1., 4., 9.]), dtype=ht.float32, device=cpu:0, split=0)
    >>> pdt_sparse.todense()
    DNDarray([[1., 0., 4.],
              [0., 0., 9.]], dtype=ht.float32, device=cpu:0, split=0)
    """
    return _operations.__binary_op_csr(torch.mul, t1, t2)


DCSR_matrix.__mul__ = lambda self, other: mul(self, other)
DCSR_matrix.__mul__.__doc__ = mul.__doc__
DCSR_matrix.__rmul__ = lambda self, other: mul(self, other)
DCSR_matrix.__rmul__.__doc__ = mul.__doc__



"""
Module Implementing the Kmedians Algorithm
"""
import heat as ht
from heat.cluster._kcluster import _KCluster
from heat.core.dndarray import DNDarray
from typing import Optional, Union, TypeVar


class KMedians(_KCluster):
    """
    K-Medians clustering algorithm [1].
    Uses the Manhattan (City-block, :math:`L_1`) metric for distance calculations

    Parameters
    ----------
    n_clusters : int, optional, default: 8
        The number of clusters to form as well as the number of centroids to generate.
    init : str or DNDarray, default: ‘random’
        Method for initialization:

             - ‘k-medians++’ : selects initial cluster centers for the clustering in a smart way to speed up convergence [2].
             - ‘random’: choose k observations (rows) at random from data for the initial centroids.
             - DNDarray: gives the initial centers, should be of Shape = (n_clusters, n_features)
    max_iter : int, default: 300
        Maximum number of iterations of the k-means algorithm for a single run.
    tol : float, default: 1e-4
        Relative tolerance with regards to inertia to declare convergence.
    random_state : int
        Determines random number generation for centroid initialization.

    References
    -------------
    [1] Hakimi, S., and O. Kariv. "An algorithmic approach to network location problems II: The p-medians." SIAM Journal on Applied Mathematics 37.3 (1979): 539-560.
    """

    def __init__(
        self,
        n_clusters: int = 8,
        init: Union[str, DNDarray] = "random",
        max_iter: int = 300,
        tol: float = 1e-4,
        random_state: int = None,
    ):
        if init == "kmedians++":
            init = "probability_based"

        super().__init__(
            metric=lambda x, y: ht.spatial.distance.manhattan(x, y, expand=True),
            n_clusters=n_clusters,
            init=init,
            max_iter=max_iter,
            tol=tol,
            random_state=random_state,
        )

    def _update_centroids(self, x: DNDarray, matching_centroids: DNDarray):
        """
        Compute coordinates of new centroid as median of the data points in ``x`` that are assigned to it

        Parameters
        ----------
        x :  DNDarray
            Input data
        matching_centroids : DNDarray
            Array filled with indeces ``i`` indicating to which cluster ``ci`` each sample point in x is assigned

        """
        new_cluster_centers = self._cluster_centers.copy()
        for i in range(self.n_clusters):
            # points in current cluster
            selection = (matching_centroids == i).astype(ht.int64)
            # Remove 0-element lines to avoid spoiling of median
            assigned_points = x * selection
            rows = assigned_points.abs().sum(axis=1) != 0
            local = assigned_points.larray[rows.larray]
            clean = ht.array(local, is_split=x.split)
            clean.balance_()
            # failsafe in case no point is assigned to this cluster
            # draw a random datapoint to continue/restart
            if clean.shape[0] == 0:
                _, displ, _ = x.comm.counts_displs_shape(shape=x.shape, axis=0)
                sample = ht.random.randint(0, x.shape[0]).item()
                proc = 0
                for p in range(x.comm.size):
                    if displ[p] > sample:
                        break
                    proc = p
                xi = ht.zeros(x.shape[1], dtype=x.dtype)
                if x.comm.rank == proc:
                    idx = sample - displ[proc]
                    xi = ht.array(x.lloc[idx, :], device=x.device, comm=x.comm)
                xi.comm.Bcast(xi, root=proc)
                new_cluster_centers[i, :] = xi
            else:
                if clean.shape[0] <= ht.MPI_WORLD.size:
                    clean.resplit_(axis=None)
                median = ht.median(clean, axis=0, keepdims=True)
                new_cluster_centers[i : i + 1, :] = median

        return new_cluster_centers

    def fit(self, x: DNDarray):
        """
        Computes the centroid of a k-medians clustering.

        Parameters
        ----------
        x : DNDarray
            Training instances to cluster. Shape = (n_samples, n_features)
        """
        # input sanitation
        if not isinstance(x, ht.DNDarray):
            raise ValueError(f"input needs to be a ht.DNDarray, but was {type(x)}")

        # initialize the clustering
        self._initialize_cluster_centers(x)
        self._n_iter = 0
        matching_centroids = ht.zeros((x.shape[0]), split=x.split, device=x.device, comm=x.comm)
        # iteratively fit the points to the centroids
        for epoch in range(self.max_iter):
            # increment the iteration count
            self._n_iter += 1
            # determine the centroids
            matching_centroids = self._assign_to_cluster(x)

            # update the centroids
            new_cluster_centers = self._update_centroids(x, matching_centroids)
            # check whether centroid movement has converged
            self._inertia = ((self._cluster_centers - new_cluster_centers) ** 2).sum()
            self._cluster_centers = new_cluster_centers.copy()
            if self.tol is not None and self._inertia <= self.tol:
                break

        self._labels = matching_centroids

        return self



"""Generalized operations for DCSR_matrix"""
import torch
import numpy as np

from heat.sparse.dcsr_matrix import DCSR_matrix

from . import factories
from ..core.communication import MPI
from ..core.dndarray import DNDarray
from ..core import types

from typing import Callable, Optional, Dict

__all__ = []


def __binary_op_csr(
    operation: Callable,
    t1: DCSR_matrix,
    t2: DCSR_matrix,
    out: Optional[DCSR_matrix] = None,
    fn_kwargs: Optional[Dict] = {},
) -> DCSR_matrix:
    """
    Generic wrapper for element-wise binary operations of two operands.
    Takes the operation function and the two operands involved in the operation as arguments.

    Parameters
    ----------
    operation : PyTorch function
        The operation to be performed. Function that performs operation elements-wise on the involved tensors,
        e.g. add values from other to self
    t1: DCSR_matrix
        The first operand involved in the operation.
    t2: DCSR_matrix
        The second operand involved in the operation.
    out: DCSR_matrix, optional
        Output buffer in which the result is placed. If not provided, a freshly allocated matrix is returned.
    fn_kwargs: Dict, optional
        keyword arguments used for the given operation
        Default: {} (empty dictionary)

    Returns
    -------
    result: ht.sparse.DCSR_matrix
        A DCSR_matrix containing the results of element-wise operation.
    """
    if not np.isscalar(t1) and not isinstance(t1, DCSR_matrix):
        raise TypeError(
            f"Only Dcsr_matrices and numeric scalars are supported, but input was {type(t1)}"
        )
    if not np.isscalar(t2) and not isinstance(t2, DCSR_matrix):
        raise TypeError(
            f"Only Dcsr_matrices and numeric scalars are supported, but input was {type(t2)}"
        )

    if not isinstance(t1, DCSR_matrix) and not isinstance(t2, DCSR_matrix):
        raise TypeError(
            f"Operator only to be used with Dcsr_matrices, but input types were {type(t1)} and {type(t2)}"
        )

    promoted_type = types.result_type(t1, t2).torch_type()

    # If one of the inputs is a scalar
    # just perform the operation on the data tensor
    # and create a new sparse matrix
    if np.isscalar(t1) or np.isscalar(t2):
        matrix = t1
        scalar = t2

        if np.isscalar(t1):
            matrix = t2
            scalar = t1

        res_values = operation(matrix.larray.values().to(promoted_type), scalar, **fn_kwargs)
        res_torch_sparse_csr = torch.sparse_csr_tensor(
            matrix.lindptr,
            matrix.lindices,
            res_values,
            size=matrix.lshape,
            device=matrix.device.torch_device,
        )
        return factories.sparse_csr_matrix(
            res_torch_sparse_csr, is_split=matrix.split, comm=matrix.comm, device=matrix.device
        )

    if t1.shape != t2.shape:
        raise ValueError(
            f"Dcsr_matrices of different shapes are not supported, but input shapes were {t1.shape} and {t2.shape}"
        )
    output_shape = t1.shape

    if t1.split is not None or t2.split is not None:
        if t1.split is None:
            t1 = factories.sparse_csr_matrix(t1.larray, split=0)

        if t2.split is None:
            t2 = factories.sparse_csr_matrix(t2.larray, split=0)

    output_split = t1.split
    output_device = t1.device
    output_comm = t1.comm
    output_balanced = t1.balanced
    output_lshape = t1.lshape

    # sanitize out buffer
    if out is not None:
        if out.shape != output_shape:
            raise ValueError(
                f"Output buffer shape is not compatible with the result. Expected {output_shape}, received {out.shape}"
            )

        if out.split != output_split:
            if out.split is None:
                out = factories.sparse_csr_matrix(out.larray, split=0)
            else:
                out = factories.sparse_csr_matrix(
                    torch.sparse_csr_tensor(
                        torch.tensor(out.indptr, dtype=torch.int64),
                        torch.tensor(out.indices, dtype=torch.int64),
                        torch.tensor(out.data),
                    )
                )

        out.device = output_device
        out.balanced = (
            output_balanced  # At this point, inputs and out buffer assumed to be balanced
        )

    # If there are no non-zero elements in either tensors, skip torch operation to
    #   1. Avoid unnecessary computation
    #   2. Avoid undefined behaviour when no data in process
    if t1.lnnz == 0 and t2.lnnz == 0:
        result = t1.larray
    else:
        result = operation(t1.larray.to(promoted_type), t2.larray.to(promoted_type), **fn_kwargs)

    if output_split is not None:
        output_gnnz = torch.tensor(result._nnz())
        output_comm.Allreduce(MPI.IN_PLACE, output_gnnz, MPI.SUM)
        output_gnnz = output_gnnz.item()
    else:
        output_gnnz = torch.tensor(result._nnz())

    output_type = types.canonical_heat_type(result.dtype)

    if out is None:
        return DCSR_matrix(
            array=torch.sparse_csr_tensor(
                result.crow_indices().to(torch.int64),
                result.col_indices().to(torch.int64),
                result.values(),
                size=output_lshape,
            ),
            gnnz=output_gnnz,
            gshape=output_shape,
            dtype=output_type,
            split=output_split,
            device=output_device,
            comm=output_comm,
            balanced=output_balanced,
        )

    out.larray.copy_(result)
    out.gnnz = output_gnnz
    out.dtype = output_type
    out.comm = output_comm
    return out



"""
Module for (pairwise) distance functions
"""
import torch
import numpy as np
from mpi4py import MPI
from typing import Callable

from ..core import factories
from ..core import types
from ..core.dndarray import DNDarray

__all__ = ["cdist", "manhattan", "rbf"]


def _euclidian(x: torch.tensor, y: torch.tensor) -> torch.tensor:
    """
    Helper function to calculate Euclidian distance between ``torch.tensors`` ``x`` and ``y``: :math:`\\sqrt(|x-y|^2)`
    Based on ``torch.cdist``. Returns 2D torch.Tensor of size :math:`m \\times n`

    Parameters
    ----------
    x : torch.Tensor
        2D tensor of size :math:`m x f`
    y : torch.Tensor
        2D tensor of size :math:`n x f`
    """
    return torch.cdist(x, y)


def _euclidian_fast(x: torch.tensor, y: torch.tensor) -> torch.tensor:
    """
    Helper function to calculate Euclidian distance between ``torch.tensors`` ``x`` and ``y``: :math:`\\sqrt(|x-y|^2)`
    Uses quadratic expansion to calculate :math:`(x-y)^2`. Returns 2D torch.Tensor of size :math:`m \\times n`

    Parameters
    ----------
    x : torch.Tensor
        2D tensor of size :math:`m x f`
    y : torch.Tensor
        2D tensor of size :math:`n x f`
    """
    return torch.sqrt(_quadratic_expand(x, y))


def _quadratic_expand(x: torch.tensor, y: torch.tensor) -> torch.tensor:
    """
    Helper function to calculate quadratic expansion :math:`|x-y|^2=|x|^2 + |y|^2 - 2xy`
    Returns 2D torch.Tensor of size :math:`m x n`

    Parameters
    ----------
    x : torch.Tensor
        2D tensor of size :math:`m x f`
    y : torch.Tensor
        2D tensor of size :math:`n x f`
    """
    x_norm = (x**2).sum(1).view(-1, 1)
    y_t = torch.transpose(y, 0, 1)
    y_norm = (y**2).sum(1).view(1, -1)

    dist = x_norm + y_norm - 2.0 * torch.mm(x, y_t)
    return torch.clamp(dist, 0.0, np.inf)


def _gaussian(x: torch.tensor, y: torch.tensor, sigma: float = 1.0) -> torch.tensor:
    """
    Helper function to calculate Gaussian distance between ``torch.Tensors`` ``x`` and ``y``: :math:`exp(-(|x-y|^2/2\\sigma^2)`
    Based on ``torch.cdist``. Returns a 2D tensor of size :math:`m x n`

    Parameters
    ----------
    x : torch.Tensor
        2D tensor of size :math:`m x f`
    y : torch.Tensor
        2D tensor of size :math:`n x f`
    sigma: float
        Scaling factor for Gaussian kernel

    """
    d2 = _euclidian(x, y) ** 2
    result = torch.exp(-d2 / (2 * sigma * sigma))
    return result


def _gaussian_fast(x: torch.tensor, y: torch.tensor, sigma: float = 1.0) -> torch.tensor:
    """
    Helper function to calculate Gaussian distance between ``torch.Tensors`` ``x`` and ``y``: :math:`exp(-(|x-y|^2/2\\sigma^2)`
    Uses quadratic expansion to calculate :math:`(x-y)^2`. Returns a 2D tensor of size :math:`m x n`

    Parameters
    ----------
    x : torch.Tensor
        2D tensor of size :math:`m x f`
    y : torch.Tensor
        2D tensor of size :math:`n x f`
    sigma: float
        Scaling factor for Gaussian kernel
    """
    d2 = _quadratic_expand(x, y)
    result = torch.exp(-d2 / (2 * sigma * sigma))
    return result


def _manhattan(x: torch.tensor, y: torch.tensor) -> torch.tensor:
    """
    Helper function to calculate Manhattan distance between ``torch.Tensors`` ``x`` and ``y``: :math:`sum(|x-y|)`
    Based on ``torch.cdist``. Returns a 2D tensor of size :math:`m x n`.

    Parameters
    ----------
    x : torch.Tensor
        2D tensor of size :math:`m x f`
    y : torch.Tensor
        2D tensor of size :math:`n x f`
    """
    return torch.cdist(x, y, p=1)


def _manhattan_fast(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    Helper function to calculate manhattan distance between ``torch.Tensors`` ``x`` and ``y``: :math:`sum(|x-y|)`
    Uses dimension expansion. Returns a 2D tensor of size :math:`m x n`.

    Parameters
    ----------
    x : torch.Tensor
        2D tensor of size :math:`m x f`
    y : torch.Tensor
        2D tensor of size :math:`n x f`
    """
    return torch.sum(torch.abs(x.unsqueeze(1) - y.unsqueeze(0)), dim=2)


def cdist(X: DNDarray, Y: DNDarray = None, quadratic_expansion: bool = False) -> DNDarray:
    """
    Calculate Euclidian distance between two DNDarrays:

    .. math:: d(x,y) = \\sqrt{(|x-y|^2)}

    Returns 2D DNDarray of size :math: `m \\times n`

    Parameters
    ----------
    X : DNDarray
        2D array of size :math: `m \\times f`
    Y : DNDarray
        2D array of size :math: `n \\times f`
    quadratic_expansion : bool
        Whether to use quadratic expansion for :math:`\\sqrt{(|x-y|^2)}` (Might yield speed-up)
    """
    if quadratic_expansion:
        return _dist(X, Y, _euclidian_fast)
    else:
        return _dist(X, Y, _euclidian)


def rbf(
    X: DNDarray, Y: DNDarray = None, sigma: float = 1.0, quadratic_expansion: bool = False
) -> DNDarray:
    """
    Calculate Gaussian distance between two DNDarrays:

    .. math:: d(x,y) = exp(-(|x-y|^2/2\\sigma^2)

    Returns 2D DNDarray of size :math: `m \\times n`

    Parameters
    ----------
    X : DNDarray
        2D array of size :math: `m \\times f`
    Y : DNDarray
        2D array of size `n \\times f`
    sigma: float
        Scaling factor for gaussian kernel
    quadratic_expansion : bool
        Whether to use quadratic expansion for :math:`\\sqrt{(|x-y|^2)}` (Might yield speed-up)
    """
    if quadratic_expansion:
        return _dist(X, Y, lambda x, y: _gaussian_fast(x, y, sigma))
    else:
        return _dist(X, Y, lambda x, y: _gaussian(x, y, sigma))


def manhattan(X: DNDarray, Y: DNDarray = None, expand: bool = False):
    """
    Calculate Manhattan distance between two DNDarrays:

    .. math:: d(x,y) = \\sum{|x_i-y_i|}

    Returns 2D DNDarray of size :math: `m \\times n`

    Parameters
    ----------
    X : DNDarray
        2D array of size :math: `m \\times f`
    Y : DNDarray
        2D array of size :math: `n \\times f`
    expand : bool
        Whether to use dimension expansion (Might yield speed-up)
    """
    if expand:
        return _dist(X, Y, lambda x, y: _manhattan_fast(x, y))
    else:
        return _dist(X, Y, lambda x, y: _manhattan(x, y))


def _dist(X: DNDarray, Y: DNDarray = None, metric: Callable = _euclidian) -> DNDarray:
    """
    Pairwise distance calculation between all elements along axis 0 of ``X`` and ``Y`` Returns 2D DNDarray of size :math: `m \\times n`
    ``X.split`` and ``Y.split`` can be distributed among axis 0.
        - if neither ``X`` nor ``Y`` is split, result will also be ``split = None \n
        - if ``X.split == 0``, result will be ``split = 0`` regardless of ``Y.split`` \n
    The distance matrix is calculated tile-wise with ring communication between the processes
    holding each a piece of ``X`` and/or ``Y``.

    Parameters
    ----------
    X : DNDarray
        2D Array of size :math: `m \\times f`
    Y : DNDarray, optional
        2D array of size `n \\times f``.
        If `Y in None, the distances will be calculated between all elements of ``X``
    metric: Callable
        The distance to be calculated between ``X`` and ``Y``
        If metric requires additional arguments, it must be handed over as a lambda function: ``lambda x, y: metric(x, y, **args)``

    Notes
    -------
    If ``X.split=None`` and ``Y.split=0``, result will be ``split=1``

    """
    if len(X.shape) > 2:
        raise NotImplementedError("Only 2D data matrices are currently supported")

    if Y is None:
        if X.dtype == types.float32:
            torch_type = torch.float32
            mpi_type = MPI.FLOAT
        elif X.dtype == types.float64:
            torch_type = torch.float64
            mpi_type = MPI.DOUBLE
        else:
            promoted_type = types.promote_types(X.dtype, types.float32)
            X = X.astype(promoted_type)
            if promoted_type == types.float32:
                torch_type = torch.float32
                mpi_type = MPI.FLOAT
            elif promoted_type == types.float64:
                torch_type = torch.float64
                mpi_type = MPI.DOUBLE
            else:
                raise NotImplementedError(f"Datatype {X.dtype} currently not supported as input")

        d = factories.zeros(
            (X.shape[0], X.shape[0]), dtype=X.dtype, split=X.split, device=X.device, comm=X.comm
        )

        if X.split is None:
            d.larray = metric(X.larray, X.larray)

        elif X.split == 0:
            comm = X.comm
            rank = comm.Get_rank()
            size = comm.Get_size()
            K, f = X.shape

            counts, displ, _ = comm.counts_displs_shape(X.shape, X.split)
            num_iter = (size + 1) // 2

            stationary = X.larray
            rows = (displ[rank], displ[rank + 1] if (rank + 1) != size else K)

            # 0th iteration, calculate diagonal
            d_ij = metric(stationary, stationary)
            d.larray[:, rows[0] : rows[1]] = d_ij
            for iter in range(1, num_iter):
                # Send rank's part of the matrix to the next process in a circular fashion
                receiver = (rank + iter) % size
                sender = (rank - iter) % size

                col1 = displ[sender]
                col2 = displ[sender + 1] if sender != size - 1 else K
                columns = (col1, col2)
                # All but the first iter processes are receiving, then sending
                if (rank // iter) != 0:
                    stat = MPI.Status()
                    comm.handle.Probe(source=sender, tag=iter, status=stat)
                    count = int(stat.Get_count(mpi_type) / f)
                    moving = torch.zeros((count, f), dtype=torch_type, device=X.device.torch_device)
                    comm.Recv(moving, source=sender, tag=iter)
                # Sending to next Process
                comm.Send(stationary, dest=receiver, tag=iter)
                # The first iter processes can now receive after sending
                if (rank // iter) == 0:
                    stat = MPI.Status()
                    comm.handle.Probe(source=sender, tag=iter, status=stat)
                    count = int(stat.Get_count(mpi_type) / f)
                    moving = torch.zeros((count, f), dtype=torch_type, device=X.device.torch_device)
                    comm.Recv(moving, source=sender, tag=iter)

                d_ij = metric(stationary, moving)
                d.larray[:, columns[0] : columns[1]] = d_ij

                # Receive calculated tile
                scol1 = displ[receiver]
                scol2 = displ[receiver + 1] if receiver != size - 1 else K
                scolumns = (scol1, scol2)
                symmetric = torch.zeros(
                    scolumns[1] - scolumns[0],
                    (rows[1] - rows[0]),
                    dtype=torch_type,
                    device=X.device.torch_device,
                )
                if (rank // iter) != 0:
                    comm.Recv(symmetric, source=receiver, tag=iter)

                # sending result back to sender of moving matrix (for symmetry)
                comm.Send(d_ij, dest=sender, tag=iter)
                if (rank // iter) == 0:
                    comm.Recv(symmetric, source=receiver, tag=iter)
                d.larray[:, scolumns[0] : scolumns[1]] = symmetric.transpose(0, 1)

            if (size + 1) % 2 != 0:  # we need one mor iteration for the first n/2 processes
                receiver = (rank + num_iter) % size
                sender = (rank - num_iter) % size
                # Case 1: only receiving
                if rank < (size // 2):
                    stat = MPI.Status()
                    comm.handle.Probe(source=sender, tag=num_iter, status=stat)
                    count = int(stat.Get_count(mpi_type) / f)
                    moving = torch.zeros((count, f), dtype=torch_type, device=X.device.torch_device)
                    comm.Recv(moving, source=sender, tag=num_iter)

                    col1 = displ[sender]
                    col2 = displ[sender + 1] if sender != size - 1 else K
                    columns = (col1, col2)

                    d_ij = metric(stationary, moving)
                    d.larray[:, columns[0] : columns[1]] = d_ij

                    # sending result back to sender of moving matrix (for symmetry)
                    comm.Send(d_ij, dest=sender, tag=num_iter)

                # Case 2 : only sending processes
                else:
                    comm.Send(stationary, dest=receiver, tag=num_iter)

                    # Receiving back result
                    scol1 = displ[receiver]
                    scol2 = displ[receiver + 1] if receiver != size - 1 else K
                    scolumns = (scol1, scol2)
                    symmetric = torch.zeros(
                        (scolumns[1] - scolumns[0], rows[1] - rows[0]),
                        dtype=torch_type,
                        device=X.device.torch_device,
                    )
                    comm.Recv(symmetric, source=receiver, tag=num_iter)
                    d.larray[:, scolumns[0] : scolumns[1]] = symmetric.transpose(0, 1)

        else:
            raise NotImplementedError(
                f"Input split was X.split = {X.split}. Splittings other than 0 or None currently not supported."
            )
    else:
        if len(Y.shape) > 2:
            raise NotImplementedError(
                f"Only 2D data matrices are supported, but input shapes were X: {X.shape}, Y: {Y.shape}"
            )

        if X.comm != Y.comm:
            raise NotImplementedError("Differing communicators not supported")

        if X.split is None:
            if Y.split is None:
                split = None
            elif Y.split == 0:
                split = 1
            else:
                raise NotImplementedError(
                    f"Input splits were X.split = {X.split}, Y.split = {Y.split}. Splittings other than 0 or None currently not supported."
                )
        elif X.split == 0:
            split = X.split
        else:
            # ToDo: Find out if even possible
            raise NotImplementedError(
                f"Input splits were X.split = {X.split}, Y.split = {Y.split}. Splittings other than 0 or None currently not supported."
            )

        promoted_type = types.promote_types(X.dtype, Y.dtype)
        promoted_type = types.promote_types(promoted_type, types.float32)
        X = X.astype(promoted_type)
        Y = Y.astype(promoted_type)
        if promoted_type == types.float32:
            torch_type = torch.float32
            mpi_type = MPI.FLOAT
        elif promoted_type == types.float64:
            torch_type = torch.float64
            mpi_type = MPI.DOUBLE
        else:
            raise NotImplementedError(f"Datatype {X.dtype} currently not supported as input")

        d = factories.zeros(
            (X.shape[0], Y.shape[0]), dtype=promoted_type, split=split, device=X.device, comm=X.comm
        )

        if X.split is None:
            d.larray = metric(X.larray, Y.larray)

        elif X.split == 0:
            if Y.split is None:
                d.larray = metric(X.larray, Y.larray)

            elif Y.split == 0:
                if X.shape[1] != Y.shape[1]:
                    raise ValueError("Inputs must have same shape[1]")

                comm = X.comm
                rank = comm.Get_rank()
                size = comm.Get_size()

                m, f = X.shape
                n = Y.shape[0]

                xcounts, xdispl, _ = X.comm.counts_displs_shape(X.shape, X.split)
                ycounts, ydispl, _ = Y.comm.counts_displs_shape(Y.shape, Y.split)
                num_iter = size

                x_ = X.larray
                stationary = Y.larray
                # rows = (xdispl[rank], xdispl[rank + 1] if (rank + 1) != size else m)
                cols = (ydispl[rank], ydispl[rank + 1] if (rank + 1) != size else n)

                # 0th iteration, calculate diagonal
                d_ij = metric(x_, stationary)
                d.larray[:, cols[0] : cols[1]] = d_ij

                for iter in range(1, num_iter):
                    # Send rank's part of the matrix to the next process in a circular fashion
                    receiver = (rank + iter) % size
                    sender = (rank - iter) % size

                    col1 = ydispl[sender]
                    col2 = ydispl[sender + 1] if sender != size - 1 else n
                    columns = (col1, col2)

                    # All but the first iter processes are receiving, then sending
                    if (rank // iter) != 0:
                        stat = MPI.Status()
                        Y.comm.handle.Probe(source=sender, tag=iter, status=stat)
                        count = int(stat.Get_count(mpi_type) / f)
                        moving = torch.zeros(
                            (count, f), dtype=torch_type, device=X.device.torch_device
                        )
                        Y.comm.Recv(moving, source=sender, tag=iter)

                    # Sending to next Process
                    Y.comm.Send(stationary, dest=receiver, tag=iter)

                    # The first iter processes can now receive after sending
                    if (rank // iter) == 0:
                        stat = MPI.Status()
                        Y.comm.handle.Probe(source=sender, tag=iter, status=stat)
                        count = int(stat.Get_count(mpi_type) / f)
                        moving = torch.zeros(
                            (count, f), dtype=torch_type, device=X.device.torch_device
                        )
                        Y.comm.Recv(moving, source=sender, tag=iter)

                    d_ij = metric(x_, moving)
                    d.larray[:, columns[0] : columns[1]] = d_ij

        elif X.split == 0:
            raise NotImplementedError(
                f"Input splits were X.split = {X.split}, Y.split = {Y.split}. Splittings other than 0 or None currently not supported."
            )
    return d



"""
distributed hierarchical SVD
"""
import numpy as np
import collections
import torch
from typing import Type, Callable, Dict, Any, TypeVar, Union, Tuple, Optional

from ..communication import MPICommunication
from ..dndarray import DNDarray
from .. import factories
from .. import types
from ..linalg import matmul, vector_norm
from ..indexing import where
from ..random import randn

from ..manipulations import vstack, hstack, diag, balance

from .. import statistics
from math import log, ceil, floor, sqrt


__all__ = ["hsvd_rank", "hsvd_rtol", "hsvd"]


#######################################################################################
# user-friendly versions of hSVD
#######################################################################################


def hsvd_rank(
    A: DNDarray,
    maxrank: int,
    compute_sv: bool = False,
    maxmergedim: Optional[int] = None,
    safetyshift: int = 5,
    silent: bool = True,
) -> Union[
    Tuple[DNDarray, DNDarray, DNDarray, float], Tuple[DNDarray, DNDarray, DNDarray], DNDarray
]:
    """
        Hierarchical SVD (hSVD) with prescribed truncation rank `maxrank`.
        If A = U diag(sigma) V^T is the true SVD of A, this routine computes an approximation for U[:,:maxrank] (and sigma[:maxrank], V[:,:maxrank]).

        The accuracy of this approximation depends on the structure of A ("low-rank" is best) and appropriate choice of parameters.

        One can expect a similar outcome from this routine as for sci-kit learn's TruncatedSVD (with `algorithm='randomized'`) although a different, determinstic algorithm is applied here. Hereby, the parameters `n_components`
        and `n_oversamples` (sci-kit learn) roughly correspond to `maxrank` and `safetyshift` (see below).

        Parameters
        ----------
        A : DNDarray
            2D-array (float32/64) of which the hSVD has to be computed.
        maxrank : int
            truncation rank. (This parameter corresponds to `n_components` in sci-kit learn's TruncatedSVD.)
        compute_sv : bool, optional
            compute_sv=True implies that also Sigma and V are computed and returned. The default is False.
        maxmergedim : int, optional
            maximal size of the concatenation matrices during the merging procedure. The default is None and results in an appropriate choice depending on the size of the local slices of A and maxrank.
            Too small choices for this parameter will result in failure if the maximal size of the concatenation matrices does not allow to merge at least two matrices. Too large choices for this parameter can cause memory errors if the resulting merging problem becomes too large.
        safetyshift : int, optional
            Increases the actual truncation rank within the computations by a safety shift. The default is 5. (There is some similarity to `n_oversamples` in sci-kit learn's TruncatedSVD.)
        silent : bool, optional
            silent=False implies that some information on the computations are printed. The default is True.

        Returns
        -------
        (Union[    Tuple[DNDarray, DNDarray, DNDarray, float], Tuple[DNDarray, DNDarray, DNDarray], DNDarray])
            if compute_sv=True: U, Sigma, V, a-posteriori error estimate for the reconstruction error ||A-U Sigma V^T ||_F / ||A||_F (computed according to [2] along the "true" merging tree).
            if compute_sv=False: U, a-posteriori error estimate

        Notes
        -------
        The size of the process local SVDs to be computed during merging is proportional to the non-split size of the input A and (maxrank + safetyshift). Therefore, conservative choice of maxrank and safetyshift is advised to avoid memory issues.
        Note that, as sci-kit learn's randomized SVD, this routine is different from `numpy.linalg.svd` because not all singular values and vectors are computed
        and even those computed may be inaccurate if the input matrix exhibts a unfavorable structure.

    See Also
    ---------
    :func:`hsvd`
    :func:`hsvd_rtol`
        References
        -------
        [1] Iwen, Ong. A distributed and incremental SVD algorithm for agglomerative data analysis on large networks. SIAM J. Matrix Anal. Appl., 37(4), 2016.
        [2] Himpe, Leibner, Rave. Hierarchical approximate proper orthogonal decomposition. SIAM J. Sci. Comput., 40 (5), 2018.
    """
    if not isinstance(A, DNDarray):
        raise TypeError("Argument needs to be a DNDarray but is {}.".format(type(A)))
    if not A.ndim == 2:
        raise ValueError("A needs to be a 2D matrix")
    if not A.dtype == types.float32 and not A.dtype == types.float64:
        raise TypeError(
            "Argument needs to be a DNDarray with datatype float32 or float64, but data type is {}.".format(
                A.dtype
            )
        )
    A_local_size = max(A.lshape_map[:, 1])

    if maxmergedim is not None and maxmergedim < 2 * (maxrank + safetyshift) + 1:
        raise RuntimeError(
            "maxmergedim=%d is too small. Please ensure `maxmergedim > 2*(maxrank + safetyshift)`, or set `maxmergedim=None` in order to work with the default value."
            % maxmergedim
        )

    if maxmergedim is None:
        if A_local_size >= 2 * (maxrank + safetyshift):
            maxmergedim = A_local_size
        else:
            maxmergedim = 2 * (maxrank + safetyshift) + 1

    return hsvd(
        A,
        maxrank=maxrank,
        maxmergedim=maxmergedim,
        rtol=None,
        safetyshift=safetyshift,
        no_of_merges=None,
        compute_sv=compute_sv,
        silent=silent,
        warnings_off=True,
    )


def hsvd_rtol(
    A: DNDarray,
    rtol: float,
    compute_sv: bool = False,
    maxrank: Optional[int] = None,
    maxmergedim: Optional[int] = None,
    safetyshift: int = 5,
    no_of_merges: Optional[int] = None,
    silent: bool = True,
) -> Union[
    Tuple[DNDarray, DNDarray, DNDarray, float], Tuple[DNDarray, DNDarray, DNDarray], DNDarray
]:
    """
        Hierchical SVD (hSVD) with prescribed upper bound on the relative reconstruction error.
        If A = U diag(sigma) V^T is the true SVD of A, this routine computes an approximation for U[:,:r] (and sigma[:r], V[:,:r])
        such that the rel. reconstruction error ||A-U[:,:r] diag(sigma[:r]) V[:,:r]^T ||_F / ||A||_F does not exceed rtol.

        The accuracy of this approximation depends on the structure of A ("low-rank" is best) and appropriate choice of parameters. This routine is similar to `hsvd_rank` with the difference that
        truncation is not performed after a fixed number (namly `maxrank` many) singular values but after such a number of singular values that suffice to capture a prescribed fraction of the amount of information
        contained in the input data (`rtol`).

        Parameters
        ----------
        A : DNDarray
            2D-array (float32/64) of which the hSVD has to be computed.
        rtol : float
            desired upper bound on the relative reconstruction error ||A-U Sigma V^T ||_F / ||A||_F. This upper bound is processed into 'local'
            tolerances during the actual computations assuming the worst case scenario of a binary "merging tree"; therefore, the a-posteriori
            error for the relative error using the true "merging tree" (see output) may be significantly smaller than rtol.
            Prescription of maxrank or maxmergedim (disabled in default) can result in loss of desired precision, but can help to avoid memory issues.
        compute_sv : bool, optional
            compute_sv=True implies that also Sigma and V are computed and returned. The default is False.
        no_of_merges : int, optional
            Maximum number of processes to be merged at each step. If no further arguments are provided (see below),
            this completely determines the "merging tree" and may cause memory issues. The default is None and results in a binary merging tree.
            Note that no_of_merges dominates maxrank and maxmergedim in the sense that at most no_of_merges processes are merged
            even if maxrank and maxmergedim would allow merging more processes.
        maxrank : int, optional
            maximal truncation rank. The default is None.
            Setting at least one of maxrank and maxmergedim is recommended to avoid memory issues, but can result in loss of desired precision.
            Setting only maxrank (and not maxmergedim) results in an appropriate default choice for maxmergedim depending on the size of the local slices of A and the value of maxrank.
        maxmergedim : int, optional
            maximal size of the concatenation matrices during the merging procedure. The default is None and results in an appropriate choice depending on the size of the local slices of A and maxrank. The default is None.
            Too small choices for this parameter will result in failure if the maximal size of the concatenation matrices does not allow to merge at least two matrices. Too large choices for this parameter can cause memory errors if the resulting merging problem becomes too large.
            Setting at least one of maxrank and maxmergedim is recommended to avoid memory issues, but can result in loss of desired precision.
            Setting only maxmergedim (and not maxrank) results in an appropriate default choice for maxrank.
        safetyshift : int, optional
            Increases the actual truncation rank within the computations by a safety shift. The default is 5.
        silent : bool, optional
            silent=False implies that some information on the computations are printed. The default is True.

        Returns
        -------
        (Union[    Tuple[DNDarray, DNDarray, DNDarray, float], Tuple[DNDarray, DNDarray, DNDarray], DNDarray])
            if compute_sv=True: U, Sigma, V, a-posteriori error estimate for the reconstruction error ||A-U Sigma V^T ||_F / ||A||_F (computed according to [2] along the "true" merging tree used in the computations).
            if compute_sv=False: U, a-posteriori error estimate

        Notes
        -------
        The maximum size of the process local SVDs to be computed during merging is proportional to the non-split size of the input A and (maxrank + safetyshift). Therefore, conservative choice of maxrank and safetyshift is advised to avoid memory issues.
        For similar reasons, prescribing only rtol and the number of processes to be merged in each step (without specifying maxrank or maxmergedim) may result in memory issues.
        Although prescribing maxrank is therefore strongly recommended to avoid memory issues, but may result in loss of desired precision (rtol). If this occures, a separate warning will be raised.

        Note that this routine is different from `numpy.linalg.svd` because not all singular values and vectors are computed and even those computed may be inaccurate if the input matrix exhibts a unfavorable structure.

        To avoid confusion, note that `rtol` in this routine does not have any similarity to `tol` in scikit learn's TruncatedSVD.
    See Also
    ---------
    :func:`hsvd`
    :func:`hsvd_rank`
        References
        -------
        [1] Iwen, Ong. A distributed and incremental SVD algorithm for agglomerative data analysis on large networks. SIAM J. Matrix Anal. Appl., 37(4), 2016.
        [2] Himpe, Leibner, Rave. Hierarchical approximate proper orthogonal decomposition. SIAM J. Sci. Comput., 40 (5), 2018.
    """
    if not isinstance(A, DNDarray):
        raise TypeError("Argument needs to be a DNDarray but is {}.".format(type(A)))
    if not A.ndim == 2:
        raise ValueError("A needs to be a 2D matrix")
    if not A.dtype == types.float32 and not A.dtype == types.float64:
        raise TypeError(
            "Argument needs to be a DNDarray with datatype float32 or float64, but data type is {}.".format(
                A.dtype
            )
        )
    A_local_size = max(A.lshape_map[:, 1])

    if maxmergedim is not None and maxrank is None:
        maxrank = floor(A_local_size / 2) - safetyshift
        if maxrank <= 0:
            raise ValueError("safetyshift is too large.")

    if maxmergedim is None and maxrank is not None:
        if A_local_size >= 2 * (maxrank + safetyshift):
            maxmergedim = A_local_size
        else:
            maxmergedim = 2 * (maxrank + safetyshift) + 1

    if (
        maxmergedim is not None
        and maxrank is not None
        and maxmergedim < 2 * (maxrank + safetyshift) + 1
    ):
        raise ValueError(
            "maxmergedim=%d is too small. Please ensure `maxmergedim > 2*(maxrank + safetyshift)`, or set `maxmergedim=None` in order to work with the default value."
            % maxmergedim
        )

    if maxmergedim is None and maxrank is None:
        if no_of_merges is None:
            no_of_merges = 2
        maxmergedim = 2 * (A.shape[1] + safetyshift) + 1
        maxrank = A.shape[1]

    if no_of_merges is not None and no_of_merges < 2:
        raise ValueError("`no_of_merges` must be >= 2.")

    return hsvd(
        A,
        maxrank=maxrank,
        maxmergedim=maxmergedim,
        rtol=rtol,
        safetyshift=safetyshift,
        no_of_merges=no_of_merges,
        compute_sv=compute_sv,
        silent=silent,
        warnings_off=True,
    )


################################################################################################
# hSVD - "full" routine for the experts
################################################################################################


def hsvd(
    A: DNDarray,
    maxrank: Optional[int] = None,
    maxmergedim: Optional[int] = None,
    rtol: Optional[float] = None,
    safetyshift: int = 0,
    no_of_merges: Optional[int] = 2,
    compute_sv: bool = False,
    silent: bool = True,
    warnings_off: bool = False,
) -> Union[
    Tuple[DNDarray, DNDarray, DNDarray, float], Tuple[DNDarray, DNDarray, DNDarray], DNDarray
]:
    """
    This function computes an approximate truncated SVD of A utilizing a distributed hiearchical algorithm; see the references.
    The present function `hsvd` is a low-level routine, provides many options/parameters, but no default values, and is not recommended for usage by non-experts since conflicts
    arising from inappropriate parameter choice will not be catched. We strongly recommend to use the corresponding high-level functions `hsvd_rank` and `hsvd_rtol` instead.

    Input
    -------
    A: DNDarray
        2D-array (float32/64) of which hSVD has to be computed
    maxrank: int, optional
        truncation rank of the SVD
    maxmergedim: int, optional
        maximal size of the concatenation matrices when "merging" the local SVDs
    rtol: float, optional
        upper bound on the relative reconstruction error ||A-U Sigma V^T ||_F / ||A||_F (may deteriorate due to other parameters)
    safetyshift: int, optional
        shift that increases the actual truncation rank of the local SVDs during the computations in order to increase accuracy
    no_of_merges: int, optional
        maximum number of local SVDs to be "merged" at one step
    compute_sv: bool, optional
        determines whether to compute U, Sigma, V (compute_sv=True) or not (then U only)
    silent: bool, optional
        determines whether to print infos on the computations performed (silent=False)
    warnings_off: bool, optional
        switch on and off warnings that are not intended for the high-level routines based on this function

    Returns
    -------
    (Union[    Tuple[DNDarray, DNDarray, DNDarray, float], Tuple[DNDarray, DNDarray, DNDarray], DNDarray])
        if compute_sv=True: U, Sigma, V, a-posteriori error estimate for the reconstruction error ||A-U Sigma V^T ||_F / ||A||_F (computed according to [2] along the "true" merging tree used in the computations).
        if compute_sv=False: U, a-posteriori error estimate

    References
    -------
    [1] Iwen, Ong. A distributed and incremental SVD algorithm for agglomerative data analysis on large networks. SIAM J. Matrix Anal. Appl., 37(4), 2016.
    [2] Himpe, Leibner, Rave. Hierarchical approximate proper orthogonal decomposition. SIAM J. Sci. Comput., 40 (5), 2018.

    See Also
    ---------
    :func:`hsvd_rank`
    :func:`hsvd_rtol`
    """
    # if split dimension is 0, transpose matrix and remember this
    transposeflag = False
    if A.split == 0:
        transposeflag = True
        A = A.T

    no_procs = A.comm.Get_size()

    Anorm = vector_norm(A)

    if rtol is not None:
        loc_atol = Anorm.larray * rtol / sqrt(2 * no_procs - 1)
    else:
        loc_atol = None

    # compute the SVDs on the 0th level
    # Important notice: in the following 'node' refers to the nodes of the tree-like merging structure of hSVD, not to the compute nodes of an HPC-cluster
    level = 0
    active_nodes = [i for i in range(no_procs)]
    if A.comm.rank == 0 and not silent:
        print(
            "hSVD level %d...\t" % level,
            "processes ",
            "\t\t".join(["%d" % an for an in active_nodes]),
        )

    U_loc, sigma_loc, err_squared_loc = compute_local_truncated_svd(
        level, A.comm.rank, A.larray, maxrank, loc_atol, safetyshift
    )
    U_loc = torch.matmul(U_loc, torch.diag(sigma_loc))

    finished = False
    while not finished:
        # communicate dimension of each nodes to all other nodes
        dims_global = A.comm.allgather(U_loc.shape[1])

        if A.comm.rank == 0 and not silent:
            print(
                "              current ranks:",
                "\t\t".join(["%d" % dims_global[an] for an in active_nodes]),
            )

        # determine future nodes and prepare sending
        future_nodes = [0]
        send_to = [[]] * no_procs
        current_idx = 0
        current_future_node = 0
        used_budget = 0
        k = 0
        counter = 0
        #        print("active_nodes", active_nodes)
        while k < len(active_nodes):
            current_idx = active_nodes[k]
            if used_budget + dims_global[current_idx] > maxmergedim or counter == no_of_merges:
                current_future_node = current_idx
                future_nodes.append(current_future_node)
                used_budget = dims_global[current_idx]
                counter = 1
            else:
                if not used_budget == 0:
                    send_to[current_idx] = current_future_node
                used_budget += dims_global[current_idx]
                counter += 1
            k += 1

        recv_from = [[]] * no_procs
        for i in future_nodes:
            recv_from[i] = [k for k in range(no_procs) if send_to[k] == i]

        if A.comm.rank in future_nodes:
            # FUTURE NODES
            # in the future nodes receive local arrays from previous level
            err_squared_loc = [err_squared_loc] + [
                torch.zeros_like(err_squared_loc) for i in recv_from[A.comm.rank]
            ]
            U_loc = [U_loc] + [
                torch.zeros(
                    (A.shape[0] + 1, dims_global[i]),
                    dtype=A.larray.dtype,
                    device=A.device.torch_device,
                )
                for i in recv_from[A.comm.rank]
            ]
            for k in range(len(recv_from[A.comm.rank])):
                # receive concatenated U_loc and err_squared_loc
                A.comm.Recv(U_loc[k + 1], recv_from[A.comm.rank][k], tag=recv_from[A.comm.rank][k])
                # separate U_loc and err_squared_loc
                err_squared_loc[k + 1] = U_loc[k + 1][-1, 0]
                U_loc[k + 1] = U_loc[k + 1][:-1]

            # concatenate the received arrays
            U_loc = torch.hstack(U_loc)
            err_squared_loc = sum(err_squared_loc)
            level += 1
            if A.comm.rank == 0 and not silent:
                print(
                    "hSVD level %d...\t" % level,
                    "processes ",
                    "\t\t".join(["%d" % fn for fn in future_nodes]),
                )
            # compute "local" SVDs on the current level

            if len(future_nodes) == 1:
                safetyshift = 0
            U_loc, sigma_loc, err_squared_loc_new = compute_local_truncated_svd(
                level, A.comm.rank, U_loc, maxrank, loc_atol, safetyshift
            )

            if len(future_nodes) > 1:
                # prepare next level or...
                U_loc = torch.matmul(U_loc, torch.diag(sigma_loc))
            err_squared_loc += err_squared_loc_new
        elif A.comm.rank in active_nodes and A.comm.rank not in future_nodes:
            # concatenate U_loc and err_squared_loc to avoid sending multiple messages
            err_squared_loc = torch.full((1, U_loc.shape[1]), err_squared_loc, device=U_loc.device)
            U_loc = torch.vstack([U_loc, err_squared_loc])
            A.comm.Send(U_loc, send_to[A.comm.rank], tag=A.comm.rank)
            # separate U_loc and err_squared_loc again
            err_squared_loc = U_loc[-1, 0]
            U_loc = U_loc[:-1]
        if len(future_nodes) == 1:
            finished = True
        else:
            active_nodes = future_nodes
    # After completion of the SVD, distribute the result from process 0 to all processes again
    # stack U_loc and err_squared_loc to avoid sending multiple messages
    err_squared_loc = torch.full((1, U_loc.shape[1]), err_squared_loc, device=U_loc.device)
    U_loc = torch.vstack([U_loc, err_squared_loc])
    U_loc_shape = A.comm.bcast(U_loc.shape, root=0)
    if A.comm.rank != 0:
        U_loc = torch.zeros(U_loc_shape, dtype=A.larray.dtype, device=A.device.torch_device)
    A.comm.Bcast(U_loc, root=0)
    # separate U_loc and err_squared_loc again
    err_squared_loc = U_loc[-1, 0]
    U = factories.array(U_loc[:-1], device=A.device, split=None, comm=A.comm)
    rel_error_estimate = (
        factories.array(err_squared_loc**0.5, device=A.device, split=None, comm=A.comm) / Anorm
    )

    # Postprocessing:
    # compute V if required or if split=0 for the input
    # in case of split=0 undo the transposition...
    if transposeflag or compute_sv:
        V = matmul(A.T, U)
        sigma = vector_norm(V, axis=0)
        if vector_norm(sigma) > 0:
            V = matmul(V, diag(1 / sigma))

        if transposeflag:
            if compute_sv:
                return V, sigma, U, rel_error_estimate
            return V, rel_error_estimate

        return U, sigma, V, rel_error_estimate

    return U, rel_error_estimate


##############################################################################################
# AUXILIARY ROUTINES
##############################################################################################


def compute_local_truncated_svd(
    level: int,
    proc_id: int,
    U_loc: torch.Tensor,
    maxrank: int,
    loc_atol: Optional[float],
    safetyshift: int,
) -> Tuple[torch.Tensor, torch.Tensor, float]:
    """
    Auxiliary routine for hsvd: computes the truncated SVD ("U-factor" and "sigma-factor" of the SVD, i.e. first and second output) of the respective local array `U_loc` together with an estimate for the truncation error (third output).
    Truncation of the SVD either to absolute (!) tolerance `loc_atol` or to maximal rank `maxrank` is performed; moreover, singular values close to or below the level of "numerical noise" (1e-14 for float64, 1e-7 for float32) are cut.
    A safetyshift is added, i.e. the final truncation rank determined from `loc_atol` and `maxrank` is increased by `safetyshift`.
    """
    U_loc, sigma_loc, _ = torch.linalg.svd(U_loc, full_matrices=False)

    if U_loc.dtype == torch.float64:
        noiselevel = 1e-14
    elif U_loc.dtype == torch.float32:
        noiselevel = 1e-7

    # the "intuitive" choice torch.argwhere is only available in torch>=1.11.0, so we need to use torch.nonzero that works similar
    no_noise_idx = torch.nonzero(sigma_loc >= noiselevel)

    if len(no_noise_idx) != 0:
        cut_noise_rank = max(no_noise_idx) + 1
        if loc_atol is None:
            loc_trunc_rank = min(maxrank, cut_noise_rank)
        else:
            # the "intuitive" choice torch.argwhere is only available in torch>=1.11.0, so we need to use torch.nonzero that works similar
            ideal_trunc_rank = min(
                torch.nonzero(
                    torch.tensor(
                        [torch.norm(sigma_loc[k:]) ** 2 for k in range(sigma_loc.shape[0] + 1)],
                        device=U_loc.device,
                    )
                    < loc_atol**2
                )
            )

            loc_trunc_rank = min(maxrank, ideal_trunc_rank, cut_noise_rank)
            if loc_trunc_rank != ideal_trunc_rank:
                print(
                    "in hSVD (level %d, process %d): abs tol = %2.2e requires truncation to rank %d, but maxrank=%d. Loss of desired precision (rtol) very likely!"
                    % (level, proc_id, loc_atol, ideal_trunc_rank, maxrank)
                )

        loc_trunc_rank = min(sigma_loc.shape[0], loc_trunc_rank + safetyshift)
        err_squared_loc = torch.linalg.norm(sigma_loc[loc_trunc_rank - safetyshift :]) ** 2
        return U_loc[:, :loc_trunc_rank], sigma_loc[:loc_trunc_rank], err_squared_loc
    else:
        err_squared_loc = torch.linalg.norm(sigma_loc) ** 2
        sigma_loc = torch.zeros(1, dtype=U_loc.dtype, device=U_loc.device)
        U_loc = torch.zeros(U_loc.shape[0], 1, dtype=U_loc.dtype, device=U_loc.device)
        return U_loc, sigma_loc, err_squared_loc



"""
file for future "full" SVD implementation
"""
from typing import Tuple
from ..dndarray import DNDarray

__all__ = ["svd"]


def svd(A: DNDarray) -> Tuple[DNDarray, DNDarray, DNDarray]:
    """
    The intended functionality is similar to `numpy.linalg.svd`, but of-course allowing for distributed-memory parallelization and GPU-support.
    """
    raise NotImplementedError(
        " Memory-distributed 'full' (i.e. non-trucated and non-approximate) SVD not implemented yet. Consider using `heat.linalg.hsvd` for an approximate, truncated  SVD instead."
    )
